from functools import cached_property
from typing import Any, Literal, Optional, TypedDict, Union

from annotated_types import Ge, Le
from pydantic import BaseModel, Field, NonNegativeInt, RootModel, computed_field, model_validator
from typing_extensions import Annotated


class TokenUsage(BaseModel):
    input_tokens: NonNegativeInt = 0
    """Number of tokens used as input to the model, inclusive of cached_input_tokens."""

    cached_input_tokens: NonNegativeInt = 0
    """Number of input tokens that were already in the KV cache."""

    forward_passes: NonNegativeInt = 0
    """Number of forward passes made to the model."""

    cached_output_tokens: NonNegativeInt = 0
    """Number of forward passes we avoided by hitting the KV cache."""

    ff_tokens: Optional[NonNegativeInt] = None
    """Number of output tokens that were fast-forwarded by the parser (if applicable)."""

    round_trips: NonNegativeInt = 0
    """Number of times a completion was generated. For remote models, this is the number of
    API calls made to the model. For local models, this is the number of times we entered a
    completion loop to generate output tokens."""

    total_latency_ms: Annotated[float, Ge(0)] = 0.0
    """Total latency of the model in milliseconds. This includes the time spent on all forward passes and
    the time spent on fast-forwarding tokens (if applicable)."""

    @computed_field  # type: ignore[misc]
    @property
    def output_tokens(self) -> NonNegativeInt:
        """Total number of output tokens generated by the model, inclusive of cached_output_tokens and ff_tokens.

        Note that this may overcount the actual number of tokens generated if some tokens were backtracked or some
        forward passes were used just to pre-fill or compute probabilities for visualization"""
        return self.forward_passes + self.cached_output_tokens + (self.ff_tokens or 0)

    @computed_field  # type: ignore[misc]
    @property
    def token_savings(self) -> Optional[Annotated[float, Ge(0), Le(1)]]:
        """The fraction of output tokens that were fast-forwarded by the parser (if applicable)."""
        if self.ff_tokens is None:
            return None
        if self.output_tokens == 0:
            return 0.0
        return self.ff_tokens / self.output_tokens

    @computed_field  # type: ignore[misc]
    @property
    def avg_latency_ms(self) -> float:
        """Average latency of tokens generated by the model."""
        if self.output_tokens == 0:
            return 0.0
        return self.total_latency_ms / self.output_tokens

    def __add__(self, other: "TokenUsage") -> "TokenUsage":
        if self.ff_tokens is None and other.ff_tokens is None:
            ff_tokens = None
        else:
            ff_tokens = (self.ff_tokens or 0) + (other.ff_tokens or 0)

        return TokenUsage(
            ff_tokens=ff_tokens,
            **{
                field: getattr(self, field) + getattr(other, field)
                for field in set(self.__class__.model_fields) - {"ff_tokens"}
            },
        )


class EngineResponse(BaseModel):
    new_bytes: bytes
    backtrack_bytes: bytes
    capture_groups: dict
    capture_group_log_probs: dict
    backtrack: NonNegativeInt = 0  # number of tokens was backtracked by the parser
    tokens: list["GenToken"] = []  # tokens associated with the generated bytes


class LegacyEngineCallResponse(BaseModel):
    new_bytes: bytes
    is_generated: bool
    new_bytes_prob: float
    capture_groups: dict
    capture_group_log_probs: dict
    new_token_count: NonNegativeInt
    backtrack: NonNegativeInt = 0  # number of tokens was backtracked by the parser
    latency_ms: NonNegativeInt = 0  # time taken by the engine to generate the output chunk
    generated_bytes: bytes = b""  # bytes generated by the engine
    generated_tokens: list["GenToken"] = []  # tokens associated with the generated bytes
    force_forwarded_bytes: bytes = b""  # bytes that were forced forwards by the parser
    force_forwarded_tokens: list["GenToken"] = []  # tokens associated with the forced forwarded bytes


class GenToken(BaseModel):
    token_id: int
    bytes: bytes
    prob: float = float("nan")
    latency_ms: float = 0.0
    is_masked: bool = False  # true if this token was ignored by the parser
    is_generated: bool = False  # true if this token was generated by the engine
    is_force_forwarded: bool = False  # true if this token was forced forwarded by the parser
    is_input: bool = False  # true if this token was part of the input
    is_backtracked: bool = False  # true if this token was backtracked by the parser


class GenTokenExtra(GenToken):
    top_k: list["GenToken"] = []


class GenData(BaseModel):
    tokens: list[int]
    mask: bytes
    temperature: float

    @computed_field  # type: ignore[misc]
    @cached_property
    def valid_next_tokens(self) -> list[int]:
        return [i for i, b in enumerate(self.mask) if b != 0]


class LLProgressCapture(BaseModel):
    object: Literal["capture"]
    name: str
    hex: str
    log_prob: float
    list_append: bool = False

    @model_validator(mode="before")
    def strip_list_append_prefix(cls, values):
        name = values["name"]
        if name.startswith("__LIST_APPEND:"):
            values["name"] = name[14:]
            # Override whatever was set
            values["list_append"] = True
        return values


class LLProgressText(BaseModel):
    object: Literal["text"]
    hex: str
    num_tokens: NonNegativeInt
    log_prob: float
    is_generated: bool


class LLProgressFinalText(BaseModel):
    object: Literal["final_text"]
    # we don't need to handle this for now


LLProgressItem = Annotated[
    Union[LLProgressCapture, LLProgressText, LLProgressFinalText],
    Field(discriminator="object"),
]


class LLProgress(RootModel):
    root: list[LLProgressItem]

    def to_engine_call_response(self) -> LegacyEngineCallResponse:
        new_bytes = b""
        new_token_count = 0
        new_bytes_prob = 0.0
        is_generated = False
        capture_groups: dict[str, Any] = {}
        capture_group_log_probs: dict[str, Any] = {}
        num_text_entries = 0

        for j in self.root:
            if isinstance(j, LLProgressCapture):
                is_generated = True
                cname = j.name
                data = bytes.fromhex(j.hex)
                if j.list_append:
                    if cname not in capture_groups or not isinstance(capture_groups[cname], list):
                        capture_groups[cname] = []
                        capture_group_log_probs[cname] = []
                    capture_groups[cname].append(data)
                    capture_group_log_probs[cname].append(j.log_prob)
                else:
                    capture_groups[cname] = data
                    capture_group_log_probs[cname] = j.log_prob
            elif isinstance(j, LLProgressText):
                # it actually should only happen once per round...
                new_bytes += bytes.fromhex(j.hex)
                new_token_count += j.num_tokens
                new_bytes_prob += j.log_prob
                is_generated |= j.is_generated
                num_text_entries += 1
        if num_text_entries > 0:
            new_bytes_prob /= num_text_entries

        return LegacyEngineCallResponse(
            new_bytes=new_bytes,
            new_token_count=new_token_count,
            new_bytes_prob=new_bytes_prob,
            is_generated=is_generated,
            capture_groups=capture_groups,
            capture_group_log_probs=capture_group_log_probs,
        )


class LLInterpreterResponse(BaseModel):
    progress: LLProgress
    stop: bool
    temperature: Optional[float]


class SamplingParams(TypedDict):
    top_p: Optional[float]
    top_k: Optional[int]
    min_p: Optional[float]
    repetition_penalty: Optional[float]



from typing import Optional

from llguidance.gbnf_to_lark import gbnf_to_lark as _gbnf_to_lark

from .._ast import GrammarNode, LarkNode, RuleNode
from .._grammar import capture, token_limit, with_temperature


def lark(
    lark_grammar: str,
    *,
    name: Optional[str] = None,
    temperature: Optional[float] = None,
    max_tokens: Optional[int] = None,
) -> GrammarNode:
    """
    Builds a guidance grammar from (a variant of) the EBNF syntax used by the Lark parsing toolkit.

    See documentation at https://github.com/guidance-ai/llguidance/blob/main/docs/syntax.md for more
    details.
    """
    node = RuleNode(name=name or "lark", value=LarkNode(lark_grammar=lark_grammar))
    if temperature is not None:
        node = with_temperature(node, temperature)
    if max_tokens is not None:
        node = token_limit(node, max_tokens)
    if name is not None:
        node = capture(node, name)

    return node


def gbnf_to_lark(gbnf_grammar: str) -> str:
    """
    Converts a GBNF (llama.cpp) grammar to Lark(-like) syntax. This is a best-effort
    conversion and may not work for all grammars. We recommend using this function
    as a starting point and then manually editing the resulting Lark grammar to suit
    your needs.

    See documentation at https://github.com/guidance-ai/llguidance/blob/main/docs/syntax.md
    for more information on the output syntax's semantics.
    """
    return _gbnf_to_lark(gbnf_grammar)



# import functions that can be called directly
# core grammar functions
from .._grammar import select, string, token_limit, with_temperature
from ._audio import audio, gen_audio

# context blocks
from ._block import block
from ._capture import capture
from ._ebnf import gbnf_to_lark, lark
from ._gen import call_tool, gen, regex
from ._image import gen_image, image
from ._json import json
from ._optional import optional
from ._role import assistant, role, system, user

# stateless library functions
from ._sequences import at_most_n_repeats, exactly_n_repeats, one_or_more, sequence, zero_or_more
from ._substring import substring
from ._tool import Tool
from ._video import gen_video, video

__all__ = [
    "Tool",
    "assistant",
    "at_most_n_repeats",
    "audio",
    "block",
    "call_tool",
    "capture",
    "exactly_n_repeats",
    "gbnf_to_lark",
    "gen",
    "gen_audio",
    "gen_image",
    "gen_video",
    "image",
    "json",
    "lark",
    "one_or_more",
    "optional",
    "regex",
    "role",
    "select",
    "sequence",
    "string",
    "substring",
    "system",
    "token_limit",
    "user",
    "video",
    "with_temperature",
    "zero_or_more",
]



"""Visualization related to trace."""

import html
import json
from typing import Optional

from ..trace import (
    ImageOutput,
    RoleCloserInput,
    RoleOpenerInput,
    TextOutput,
    TokenOutput,
    TraceHandler,
    TraceNode,
)


def trace_node_to_html(node: TraceNode, prettify_roles=False) -> str:
    """Represents trace path as html string.

    Args:
        node: Trace node that designates the end of a trace path for HTML output.
        prettify_roles: Enables prettier formatting for roles.

    Returns:
        HTML string of trace path as html.
    """
    buffer = []
    node_path = list(node.path())
    active_role: Optional[TraceNode] = None

    for node in node_path:
        # Check if any input is a role opener or closer
        for input_attr in node.input:
            if isinstance(input_attr, RoleOpenerInput):
                active_role = node
                break
            elif isinstance(input_attr, RoleCloserInput):
                active_role = node
                break
        for output_attr in node.output:
            if isinstance(output_attr, TextOutput):
                if active_role is not None:
                    # Find the first RoleOpenerInput in the active role's input list
                    role_opener_input = next(
                        (inp for inp in active_role.input if isinstance(inp, RoleOpenerInput)), None
                    )
                    if (
                        prettify_roles
                        and role_opener_input is not None
                        and (role_name := role_opener_input.name) is not None
                    ):
                        fmt = f"<div style='display: flex; border-bottom: 1px solid rgba(127, 127, 127, 0.2);  justify-content: center; align-items: center;'><div style='flex: 0 0 80px; opacity: 0.5;'>{role_name.lower()}</div><div style='flex-grow: 1; padding: 5px; padding-top: 10px; padding-bottom: 10px; margin-top: 0px; white-space: pre-wrap; margin-bottom: 0px;'>"
                        buffer.append(fmt)
                    if not prettify_roles:
                        buffer.append("<span style='background-color: rgba(255, 180, 0, 0.3); border-radius: 3px;'>")

                if not (active_role and prettify_roles):
                    attr = output_attr
                    latency = f"{attr.latency_ms:.2f}"
                    chunk_text = attr.value

                if not isinstance(attr, TokenOutput):
                    if attr.is_generated:
                        fmt = f"<span style='background-color: rgba({0}, {255}, {0}, 0.15); border-radius: 3ps;' title='Chunk: {chunk_text}\nlatency_ms: {latency}'>{html.escape(chunk_text)}</span>"
                    elif attr.is_force_forwarded:
                        fmt = f"<span style='background-color: rgba({0}, {0}, {255}, 0.15); border-radius: 3ps;' title='Chunk: {chunk_text}\nlatency_ms: {latency}'>{html.escape(chunk_text)}</span>"
                    else:
                        fmt = f"<span style='background-color: rgba({255}, {255}, {255}, 0.15); border-radius: 3ps;' title='Chunk: {chunk_text}\nlatency_ms: {latency}'>{html.escape(chunk_text)}</span>"
                else:
                    token = attr.token
                    token_str = token.token
                    # assert token_str == chunk_text

                    prob = token.prob  # TODO: what if nan?
                    top_k: dict[str, str] = {}
                    # find the correct token
                    for _token in attr.top_k or []:
                        top_k[f"{_token.token}"] = f"{_token.prob} - Masked: {_token.masked}"
                    top_k_repr = json.dumps(top_k, indent=2)

                    if attr.is_generated:
                        fmt = f"<span style='background-color: rgba({0}, {127 + int(127 * prob)}, {0}, 0.15); border-radius: 3ps;' title='Token: \"{token_str}\" : {prob}\nTop-k: {top_k_repr}\nlatency_ms: {latency}'>{html.escape(token_str)}</span>"
                    elif attr.is_force_forwarded:
                        fmt = f"<span style='background-color: rgba({0}, {0}, {127 + int(127 * prob)}, 0.15); border-radius: 3ps;' title='Token: \"{token_str}\" : {prob}\nTop-k: {top_k_repr}\nlatency_ms: {latency}'>{html.escape(token_str)}</span>"
                    else:
                        fmt = f"<span style='background-color: rgba({255}, {255}, {255}, 0.15); border-radius: 3ps;' title='Token: \"{token_str}\" : {prob}\nTop-k: {top_k_repr}'>{html.escape(token_str)}</span>"

                buffer.append(fmt)

            if active_role is not None:
                if not prettify_roles:
                    buffer.append("</span>")
                # Check if any input in active role is a RoleCloserInput
                has_role_closer = any(isinstance(inp, RoleCloserInput) for inp in active_role.input)
                if has_role_closer and prettify_roles:
                    buffer.append(f"</div></div>")
                active_role = None
            elif isinstance(output_attr, ImageOutput):
                buffer.append(
                    f'<img src="data:image/png;base64,{output_attr.value.decode()}" style="max-width" 400px; vertical-align: middle; margin: 4px;">'
                )

    buffer.insert(
        0,
        "<pre style='margin: 0px; padding: 0px; vertical-align: middle; padding-left: 8px; margin-left: -8px; border-radius: 0px; border-left: 1px solid rgba(127, 127, 127, 0.2); white-space: pre-wrap; font-family: ColfaxAI, Arial; font-size: 15px; line-height: 23px;'>",
    )
    buffer.append("</pre>")
    return "".join(buffer)


def trace_node_to_str(node: TraceNode) -> str:
    """Visualize output attributes of a trace node up to the root.

    Users should not be accessing this function. For debugging purposes.

    Args:
        node: The trace node to visualize.
    Returns:
        Output as string.
    """
    buffer = []
    for node in node.path():
        for output_attr in node.output:
            if isinstance(output_attr, TextOutput):
                buffer.append(str(output_attr))
    return "".join(buffer)


def display_trace_tree(trace_handler: TraceHandler) -> None:
    """Visualize tree of a trace node going down to all its leaves.

    Users should not normally be accessing this function. For debugging purposes.

    Args:
        trace_handler: Trace handler needed to pull user-defined identifiers of trace nodes.
    """
    from anytree import Node, RenderTree  # type: ignore[import-untyped]

    root = trace_handler.root()
    trace_viz_map: dict[TraceNode, Node] = {}
    for node in root.traverse(bfs=False):
        viz_parent = trace_viz_map.get(node.parent, None)
        viz_node = Node(f"{trace_handler.node_id_map[node]}:{node!r}", parent=viz_parent)
        trace_viz_map[node] = viz_node
    viz_root = trace_viz_map[root]

    for pre, fill, node in RenderTree(viz_root):
        tree_str = "%s%s" % (pre, node.name)
        print(tree_str)



import dataclasses
import re
from typing import Optional, Sequence, Union

from ._ast import (
    Function,
    GrammarNode,
    LiteralNode,
    RegexNode,
    RepeatNode,
    RuleNode,
    SelectNode,
    SubgrammarNode,
    _parse_tags,
)


def string(s: str) -> LiteralNode:
    return LiteralNode(s)


def regex(pattern: str) -> RegexNode:
    return RegexNode(pattern)


def gen(
    regex: Optional[str] = None,
    stop: Optional[str] = None,
    stop_regex: Optional[str] = None,
    suffix: Optional[str] = None,
    stop_capture: Optional[str] = None,
    name: Optional[str] = None,
    temperature: Optional[float] = None,
    max_tokens: Optional[int] = None,
    list_append: bool = False,
) -> RuleNode:
    if stop is not None and stop_regex is not None:
        raise ValueError("You cannot specify both a stop and a stop_regex")

    stop_value: Union[LiteralNode, RegexNode, None] = None
    if stop is not None:
        stop_value = LiteralNode(stop)
    elif stop_regex is not None:
        stop_value = RegexNode(stop_regex)

    node = RuleNode(
        name=name or "gen",
        value=RegexNode(regex),
        capture=name,
        stop=stop_value,
        suffix=LiteralNode(suffix) if suffix else None,
        stop_capture=stop_capture,
        list_append=list_append,
        temperature=temperature,
        max_tokens=max_tokens,
    )
    return node


def select(
    options: Sequence[Union[str, int, float, GrammarNode]],
    name: Optional[str] = None,
    list_append: bool = False,
) -> GrammarNode:
    """Choose between a set of options.

    This function constrains the next generation from the LLM to be one of the
    given `options`.
    If the list only has a single element, then that value can be returned
    immediately, without calling the LLM.

        >>> lm += select(["Temeraire", "Redoutable", "Bucentaure"], name="my_selection")
        >>> print(lm["my_selection"])
        Temeraire

    Parameters
    ----------
    name : str or None
        If this is not None then the the results of the generation will be saved as a variable on
        the Model object (so you can access the result as `lm["var_name"]`).

    options : list
        The set of available choices for the next generation

    list_append : bool
        If this is True then the results saved to `lm[name]` will not be written directly but rather appended
        to a list (if no list with the current name is present one will be created). This is useful for
        building lists inside python loops.
    """
    alternatives: list[GrammarNode] = []
    for v in options:
        if isinstance(v, (int, float)):
            alternatives.append(string(str(v)))
        elif isinstance(v, str):
            node = _parse_tags(v)
            if isinstance(node, Function):
                raise ValueError("You cannot select between stateful functions in the current guidance implementation!")
            if callable(node):
                raise ValueError(
                    "Did you pass a function without calling it to select? You need to pass the results of a called guidance function to select."
                )
            alternatives.append(node)
        elif isinstance(v, GrammarNode):
            alternatives.append(v)
        else:
            raise ValueError(f"Option {v} is not a valid type: {type(v)}")

    return RuleNode(
        name=name or "select",
        value=SelectNode(tuple(alternatives)),
        capture=name,
        list_append=list_append,
    )


def repeat(value: Union[str, int, float, GrammarNode], min: int, max: Optional[int] = None) -> GrammarNode:
    node: GrammarNode
    if isinstance(value, (int, float)):
        node = string(str(value))
    elif isinstance(value, str):
        _node = _parse_tags(value)
        if isinstance(_node, Function):
            raise ValueError("You cannot repeat a stateful function in the current guidance implementation!")
        if callable(_node):
            raise ValueError(
                "Did you pass a function without calling it? You need to pass the results of a called guidance function to repeat."
            )
        node = _node
    elif isinstance(value, GrammarNode):
        node = value
    else:
        raise ValueError(f"Value {value} is not a valid type: {type(value)}")

    return RuleNode(
        name="repeat",
        value=RepeatNode(node, min, max),
    )


def token_limit(value: GrammarNode, max_tokens: int) -> RuleNode:
    """This sets the token limit to be used for the given portion of the grammar."""

    def inner(value: GrammarNode) -> RuleNode:
        if isinstance(value, RuleNode):
            return dataclasses.replace(value, max_tokens=max_tokens)
        else:
            return RuleNode(name="token_limit", value=value, max_tokens=max_tokens)

    try:
        return inner(value)
    except ValueError:
        return inner(subgrammar(value))


def with_temperature(value: GrammarNode, temperature: float) -> RuleNode:
    """This sets the sampling temperature to be used for the given portion of the grammar.

    Note that if the grammar passed to us already has some portions with a temperature
    setting in place, those settings will not be overridden.
    """

    def inner(value: GrammarNode) -> RuleNode:
        if isinstance(value, RuleNode):
            return dataclasses.replace(value, temperature=temperature)
        else:
            return RuleNode(name="with_temperature", value=value, temperature=temperature)

    try:
        return inner(value)
    except ValueError:
        return inner(subgrammar(value))


def capture(value: GrammarNode, name: str, list_append: bool = False) -> RuleNode:
    if isinstance(value, RuleNode) and value.capture is None:
        return dataclasses.replace(value, capture=name, list_append=list_append)
    else:
        return RuleNode(name="capture", value=value, capture=name, list_append=list_append)


def subgrammar(
    body: GrammarNode,
    name: Optional[str] = None,
    skip_regex: Optional[str] = None,
    max_tokens: Optional[int] = None,
    temperature: Optional[float] = None,
) -> RuleNode:
    capture_name = name
    name = name or (body.name if isinstance(body, RuleNode) else "subgrammar")
    node = RuleNode(name=name or "subgrammar", value=SubgrammarNode(body=body, skip_regex=skip_regex))
    if max_tokens:
        node = token_limit(node, max_tokens)
    if temperature:
        node = with_temperature(node, temperature)
    if capture_name:
        node = capture(node, capture_name)
    return node


def quote_regex(value: str) -> str:
    return re.sub(r"([\\+*?^$(){}\[\]\.|])", r"\\\1", value)



import inspect
import warnings
from typing import Union


class ChatTemplate:
    """Contains template for all chat and instruct tuned models."""

    def get_role_start(self, role_name: str, **kwargs):
        raise NotImplementedError("You need to use a ChatTemplate subclass that overrides the get_role_start method")

    def get_role_end(self, role_name: Union[str, None] = None):
        raise NotImplementedError("You need to use a ChatTemplate subclass that overrides the get_role_start method")


class ChatTemplateCache:
    def __init__(self) -> None:
        self._cache: dict[str, ChatTemplate] = {}

    def __getitem__(self, key: str) -> ChatTemplate:
        key_compact = key.replace(" ", "")
        return self._cache[key_compact]

    def __setitem__(self, key: str, value):
        key_compact = key.replace(" ", "")
        self._cache[key_compact] = value

    def __contains__(self, key: str):
        key_compact = key.replace(" ", "")
        return key_compact in self._cache


# Feels weird having to instantiate this, but it's a singleton for all purposes
# TODO [HN]: Add an alias system so we can instantiate with other simple keys (e.g. "llama2" instead of the full template string)
CHAT_TEMPLATE_CACHE = ChatTemplateCache()


class UnsupportedRoleException(Exception):
    def __init__(self, role_name, instance):
        self.role_name = role_name
        self.instance = instance
        super().__init__(self._format_message())

    def _format_message(self):
        return f"Role {self.role_name} is not supported by the {self.instance.__class__.__name__} chat template. "


def load_template_class(chat_template=None):
    """Utility method to find the best chat template.

    Order of precedence:
    - If it's a chat template class, use it directly
    - If it's a string, check the cache of popular model templates
    - If it's a string and not in the cache, try to create a class dynamically
    - [TODO] If it's a string and can't be created, default to ChatML and raise a warning
    - If it's None, default to ChatML and raise a warning
    """
    if inspect.isclass(chat_template) and issubclass(chat_template, ChatTemplate):
        if chat_template is ChatTemplate:
            raise Exception("You can't use the base ChatTemplate class directly. Create or use a subclass instead.")
        return chat_template

    elif isinstance(chat_template, str):
        # First check the cache of popular model types
        # TODO: Expand keys of cache to include aliases for popular model types (e.g. "llama2, phi3")
        # Can possibly accomplish this with an "aliases" dictionary that maps all aliases to the canonical key in cache
        if chat_template in CHAT_TEMPLATE_CACHE:
            return CHAT_TEMPLATE_CACHE[chat_template]
        # TODO: Add logic here to try to auto-create class dynamically via _template_class_from_string method

    # Only warn when a user provided a chat template that we couldn't load
    if chat_template is not None:
        warnings.warn(
            f"""Chat template {chat_template} was unable to be loaded directly into guidance.
                        Defaulting to the ChatML format which may not be optimal for the selected model. 
                        For best results, create and pass in a `guidance.ChatTemplate` subclass for your model."""
        )

    # By default, use the ChatML Template. Warnings to user will happen downstream only if they use chat roles.
    return ChatMLTemplate


def _template_class_from_string(template_str):
    """Utility method to try to create a chat template class from a string."""
    # TODO: Try to build this, perhaps based on passing unit tests we create?
    pass


# CACHE IMPLEMENTATIONS:

# --------------------------------------------------
# @@@@ ChatML @@@@
# --------------------------------------------------
# Note that all grammarless models will default to this syntax, since we typically send chat formatted messages.
chatml_template = "{% for message in messages %}{{'<|im_start|>' + message['role'] + '\n' + message['content'] + '<|im_end|>' + '\n'}}{% endfor %}"


class ChatMLTemplate(ChatTemplate):
    template_str = chatml_template

    def get_role_start(self, role_name):
        return f"<|im_start|>{role_name}\n"

    def get_role_end(self, role_name=None):  # noqa ARG002
        return "<|im_end|>\n"


CHAT_TEMPLATE_CACHE[chatml_template] = ChatMLTemplate


# --------------------------------------------------
# @@@@ Llama-2 @@@@
# --------------------------------------------------
# [05/08/24] https://huggingface.co/meta-llama/Llama-2-7b-chat-hf/blob/main/tokenizer_config.json#L12
llama2_template = "{% if messages[0]['role'] == 'system' %}{% set loop_messages = messages[1:] %}{% set system_message = messages[0]['content'] %}{% else %}{% set loop_messages = messages %}{% set system_message = false %}{% endif %}{% for message in loop_messages %}{% if (message['role'] == 'user') != (loop.index0 % 2 == 0) %}{{ raise_exception('Conversation roles must alternate user/assistant/user/assistant/...') }}{% endif %}{% if loop.index0 == 0 and system_message != false %}{% set content = '<<SYS>>\\n' + system_message + '\\n<</SYS>>\\n\\n' + message['content'] %}{% else %}{% set content = message['content'] %}{% endif %}{% if message['role'] == 'user' %}{{ bos_token + '[INST] ' + content.strip() + ' [/INST]' }}{% elif message['role'] == 'assistant' %}{{ ' '  + content.strip() + ' ' + eos_token }}{% endif %}{% endfor %}"


class Llama2ChatTemplate(ChatTemplate):
    # available_roles = ["system", "user", "assistant"]
    template_str = llama2_template

    def get_role_start(self, role_name):
        if role_name == "system":
            return "[INST] <<SYS>>\n"
        elif role_name == "user":
            return "<s>[INST] "
        elif role_name == "assistant":
            return " "
        else:
            raise UnsupportedRoleException(role_name, self)

    def get_role_end(self, role_name=None):
        if role_name == "system":
            return "\n<</SYS>>"
        elif role_name == "user":
            return " [/INST]"
        elif role_name == "assistant":
            return " </s>"
        else:
            raise UnsupportedRoleException(role_name, self)


CHAT_TEMPLATE_CACHE[llama2_template] = Llama2ChatTemplate


# --------------------------------------------------
# @@@@ Llama-3 @@@@
# --------------------------------------------------
# [05/08/24] https://huggingface.co/meta-llama/Meta-Llama-3-8B-Instruct/blob/main/tokenizer_config.json#L2053
llama3_template = "{% set loop_messages = messages %}{% for message in loop_messages %}{% set content = '<|start_header_id|>' + message['role'] + '<|end_header_id|>\n\n'+ message['content'] | trim + '<|eot_id|>' %}{% if loop.index0 == 0 %}{% set content = bos_token + content %}{% endif %}{{ content }}{% endfor %}{% if add_generation_prompt %}{{ '<|start_header_id|>assistant<|end_header_id|>\n\n' }}{% endif %}"


class Llama3ChatTemplate(ChatTemplate):
    # available_roles = ["system", "user", "assistant"]
    template_str = llama3_template

    def get_role_start(self, role_name):
        if role_name == "system":
            return "<|begin_of_text|><|start_header_id|>system<|end_header_id|>\n\n"
        elif role_name == "user":
            return "<|start_header_id|>user<|end_header_id|>\n\n"
        elif role_name == "assistant":
            return "<|start_header_id|>assistant<|end_header_id|>\n\n"
        else:
            raise UnsupportedRoleException(role_name, self)

    def get_role_end(self, role_name=None):  # noqa ARG002
        return "<|eot_id|>"


CHAT_TEMPLATE_CACHE[llama3_template] = Llama3ChatTemplate

# --------------------------------------------------
# @@@@ Phi-3 @@@@
# --------------------------------------------------
# [05/08/24] https://huggingface.co/microsoft/Phi-3-mini-4k-instruct/blob/main/tokenizer_config.json#L119
phi3_mini_template = "{% for message in messages %}{% if message['role'] == 'system' %}{{'<|system|>\n' + message['content'] + '<|end|>\n'}}{% elif message['role'] == 'user' %}{{'<|user|>\n' + message['content'] + '<|end|>\n'}}{% elif message['role'] == 'assistant' %}{{'<|assistant|>\n' + message['content'] + '<|end|>\n'}}{% endif %}{% endfor %}{% if add_generation_prompt %}{{ '<|assistant|>\n' }}{% else %}{{ eos_token }}{% endif %}"
phi3_mini_llamacpp_template = "{{ bos_token }}{% for message in messages %}{% if (message['role'] == 'user') %}{{'<|user|>' + '\n' + message['content'] + '<|end|>' + '\n' + '<|assistant|>' + '\n'}}{% elif (message['role'] == 'assistant') %}{{message['content'] + '<|end|>' + '\n'}}{% endif %}{% endfor %}"


class Phi3MiniChatTemplate(ChatTemplate):
    # available_roles = ["user", "assistant"]
    template_str = phi3_mini_template

    def get_role_start(self, role_name):
        if role_name == "user":
            return "<|user|>\n"
        elif role_name == "assistant":
            return "<|assistant|>\n"
        elif role_name == "system":
            return "<|system|>\n"
        else:
            raise UnsupportedRoleException(role_name, self)

    def get_role_end(self, role_name=None):  # noqa ARG002
        return "<|end|>\n"


CHAT_TEMPLATE_CACHE[phi3_mini_template] = Phi3MiniChatTemplate
CHAT_TEMPLATE_CACHE[phi3_mini_llamacpp_template] = Phi3MiniChatTemplate

# https://huggingface.co/microsoft/Phi-3-small-8k-instruct/blob/main/tokenizer_config.json
phi3_small_template = "{{ bos_token }}{% for message in messages %}{{'<|' + message['role'] + '|>' + '\n' + message['content'] + '<|end|>\n' }}{% endfor %}{% if add_generation_prompt %}{{ '<|assistant|>\n' }}{% else %}{{ eos_token }}{% endif %}"


# https://huggingface.co/microsoft/Phi-3-medium-4k-instruct/blob/main/tokenizer_config.json#L119
phi3_medium_template = "{% for message in messages %}{% if (message['role'] == 'user') %}{{'<|user|>' + '\n' + message['content'] + '<|end|>' + '\n' + '<|assistant|>' + '\n'}}{% elif (message['role'] == 'assistant') %}{{message['content'] + '<|end|>' + '\n'}}{% endif %}{% endfor %}"


# Although the templates are different, the roles are the same between medium and small (for now)
class Phi3SmallMediumChatTemplate(ChatTemplate):
    # available_roles = ["user", "assistant"]
    template_str = phi3_small_template

    def get_role_start(self, role_name):
        if role_name == "user":
            return "<|user|>\n"
        elif role_name == "assistant":
            return "<|assistant|>\n"
        else:
            raise UnsupportedRoleException(role_name, self)

    def get_role_end(self, role_name=None):  # noqa ARG002
        return "<|end|>\n"


CHAT_TEMPLATE_CACHE[phi3_small_template] = Phi3SmallMediumChatTemplate
CHAT_TEMPLATE_CACHE[phi3_medium_template] = Phi3SmallMediumChatTemplate

# --------------------------------------------------
# @@@@ Phi-4-mini-instruct @@@@
# --------------------------------------------------
# [06/16/25] https://huggingface.co/microsoft/Phi-4-mini-instruct/blob/main/tokenizer_config.json#L104

phi_4_mini_template = "{% for message in messages %}{% if message['role'] == 'system' and 'tools' in message and message['tools'] is not none %}{{ '<|' + message['role'] + '|>' + message['content'] + '<|tool|>' + message['tools'] + '<|/tool|>' + '<|end|>' }}{% else %}{{ '<|' + message['role'] + '|>' + message['content'] + '<|end|>' }}{% endif %}{% endfor %}{% if add_generation_prompt %}{{ '<|assistant|>' }}{% else %}{{ eos_token }}{% endif %}"


class Phi4MiniChatTemplate(ChatTemplate):
    # available_roles = ["user", "assistant", "system"]
    template_str = phi_4_mini_template

    def get_role_start(self, role_name):
        return f"<|{role_name}|>"

    def get_role_end(self, role_name=None):  # noqa ARG002
        return "<|end|>"


CHAT_TEMPLATE_CACHE[phi_4_mini_template] = Phi4MiniChatTemplate

# --------------------------------------------------
# @@@@ Mistral-7B-Instruct-v0.2 @@@@
# --------------------------------------------------
# [05/08/24] https://huggingface.co/mistralai/Mistral-7B-Instruct-v0.2/blob/main/tokenizer_config.json#L42
mistral_7b_instruct_template = "{%- if messages[0]['role'] == 'system' %}\n    {%- set system_message = messages[0]['content'] %}\n    {%- set loop_messages = messages[1:] %}\n{%- else %}\n    {%- set loop_messages = messages %}\n{%- endif %}\n\n{{- bos_token }}\n{%- for message in loop_messages %}\n    {%- if (message['role'] == 'user') != (loop.index0 % 2 == 0) %}\n        {{- raise_exception('After the optional system message, conversation roles must alternate user/assistant/user/assistant/...') }}\n    {%- endif %}\n    {%- if message['role'] == 'user' %}\n        {%- if loop.first and system_message is defined %}\n            {{- ' [INST] ' + system_message + '\\n\\n' + message['content'] + ' [/INST]' }}\n        {%- else %}\n            {{- ' [INST] ' + message['content'] + ' [/INST]' }}\n        {%- endif %}\n    {%- elif message['role'] == 'assistant' %}\n        {{- ' ' + message['content'] + eos_token}}\n    {%- else %}\n        {{- raise_exception('Only user and assistant roles are supported, with the exception of an initial optional system message!') }}\n    {%- endif %}\n{%- endfor %}\n"
mistral_7b_instruct_llamacpp_template = "{{ bos_token }}{% for message in messages %}{% if (message['role'] == 'user') != (loop.index0 % 2 == 0) %}{{ raise_exception('Conversation roles must alternate user/assistant/user/assistant/...') }}{% endif %}{% if message['role'] == 'user' %}{{ '[INST] ' + message['content'] + ' [/INST]' }}{% elif message['role'] == 'assistant' %}{{ message['content'] + eos_token}}{% else %}{{ raise_exception('Only user and assistant roles are supported!') }}{% endif %}{% endfor %}"


class Mistral7BInstructChatTemplate(ChatTemplate):
    # available_roles = ["user", "assistant"]
    template_str = mistral_7b_instruct_template

    def get_role_start(self, role_name):
        if role_name == "user":
            return " [INST] "
        elif role_name == "assistant":
            return " "
        elif role_name == "system":
            raise ValueError("Please include system instructions in the first user message")
        else:
            raise UnsupportedRoleException(role_name, self)

    def get_role_end(self, role_name=None):
        if role_name == "user":
            return " [/INST]"
        elif role_name == "assistant":
            return "</s>"
        else:
            raise UnsupportedRoleException(role_name, self)


CHAT_TEMPLATE_CACHE[mistral_7b_instruct_template] = Mistral7BInstructChatTemplate
CHAT_TEMPLATE_CACHE[mistral_7b_instruct_llamacpp_template] = Mistral7BInstructChatTemplate

# --------------------------------------------------
# @@@@ Gemma-2-9b-it @@@@
# --------------------------------------------------
# From https://huggingface.co/google/gemma-2-9b-it/blob/main/tokenizer_config.json#L1747
gemma2_9b_it_template = "{{ bos_token }}{% if messages[0]['role'] == 'system' %}{{ raise_exception('System role not supported') }}{% endif %}{% for message in messages %}{% if (message['role'] == 'user') != (loop.index0 % 2 == 0) %}{{ raise_exception('Conversation roles must alternate user/assistant/user/assistant/...') }}{% endif %}{% if (message['role'] == 'assistant') %}{% set role = 'model' %}{% else %}{% set role = message['role'] %}{% endif %}{{ '<start_of_turn>' + role + '\n' + message['content'] | trim + '<end_of_turn>\n' }}{% endfor %}{% if add_generation_prompt %}{{'<start_of_turn>model\n'}}{% endif %}"


class Gemma29BInstructChatTemplate(ChatTemplate):
    # available_roles = ["user", "assistant"]
    template_str = gemma2_9b_it_template

    def get_role_start(self, role_name):
        if role_name == "user":
            return "<start_of_turn>user\n"
        elif role_name == "assistant":
            return "<start_of_turn>model\n"
        elif role_name == "system":
            raise ValueError("System Role not supported")
        else:
            raise UnsupportedRoleException(role_name, self)

    def get_role_end(self, role_name=None):
        if role_name == "user":
            return "<end_of_turn>\n"
        elif role_name == "assistant":
            return "<end_of_turn>\n"
        else:
            raise UnsupportedRoleException(role_name, self)


CHAT_TEMPLATE_CACHE[gemma2_9b_it_template] = Gemma29BInstructChatTemplate

# --------------------------------------------------
# @@@@ Qwen2.5 @@@@
# --------------------------------------------------
# From https://huggingface.co/Qwen/Qwen2.5-0.5B/blob/060db6499f32faf8b98477b0a26969ef7d8b9987/tokenizer_config.json#L198
qwen2dot5_template = "{%- if tools %}\n    {{- '<|im_start|>system\\n' }}\n    {%- if messages[0]['role'] == 'system' %}\n        {{- messages[0]['content'] }}\n    {%- else %}\n        {{- 'You are a helpful assistant.' }}\n    {%- endif %}\n    {{- \"\\n\\n# Tools\\n\\nYou may call one or more functions to assist with the user query.\\n\\nYou are provided with function signatures within <tools></tools> XML tags:\\n<tools>\" }}\n    {%- for tool in tools %}\n        {{- \"\\n\" }}\n        {{- tool | tojson }}\n    {%- endfor %}\n    {{- \"\\n</tools>\\n\\nFor each function call, return a json object with function name and arguments within <tool_call></tool_call> XML tags:\\n<tool_call>\\n{\\\"name\\\": <function-name>, \\\"arguments\\\": <args-json-object>}\\n</tool_call><|im_end|>\\n\" }}\n{%- else %}\n    {%- if messages[0]['role'] == 'system' %}\n        {{- '<|im_start|>system\\n' + messages[0]['content'] + '<|im_end|>\\n' }}\n    {%- else %}\n        {{- '<|im_start|>system\\nYou are a helpful assistant.<|im_end|>\\n' }}\n    {%- endif %}\n{%- endif %}\n{%- for message in messages %}\n    {%- if (message.role == \"user\") or (message.role == \"system\" and not loop.first) or (message.role == \"assistant\" and not message.tool_calls) %}\n        {{- '<|im_start|>' + message.role + '\\n' + message.content + '<|im_end|>' + '\\n' }}\n    {%- elif message.role == \"assistant\" %}\n        {{- '<|im_start|>' + message.role }}\n        {%- if message.content %}\n            {{- '\\n' + message.content }}\n        {%- endif %}\n        {%- for tool_call in message.tool_calls %}\n            {%- if tool_call.function is defined %}\n                {%- set tool_call = tool_call.function %}\n            {%- endif %}\n            {{- '\\n<tool_call>\\n{\"name\": \"' }}\n            {{- tool_call.name }}\n            {{- '\", \"arguments\": ' }}\n            {{- tool_call.arguments | tojson }}\n            {{- '}\\n</tool_call>' }}\n        {%- endfor %}\n        {{- '<|im_end|>\\n' }}\n    {%- elif message.role == \"tool\" %}\n        {%- if (loop.index0 == 0) or (messages[loop.index0 - 1].role != \"tool\") %}\n            {{- '<|im_start|>user' }}\n        {%- endif %}\n        {{- '\\n<tool_response>\\n' }}\n        {{- message.content }}\n        {{- '\\n</tool_response>' }}\n        {%- if loop.last or (messages[loop.index0 + 1].role != \"tool\") %}\n            {{- '<|im_end|>\\n' }}\n        {%- endif %}\n    {%- endif %}\n{%- endfor %}\n{%- if add_generation_prompt %}\n    {{- '<|im_start|>assistant\\n' }}\n{%- endif %}\n"

# From https://huggingface.co/Qwen/Qwen2.5-0.5B-Instruct/blob/7ae557604adf67be50417f59c2c2f167def9a775/tokenizer_config.json#L198
qwen2dot5_it_template = "{%- if tools %}\n    {{- '<|im_start|>system\\n' }}\n    {%- if messages[0]['role'] == 'system' %}\n        {{- messages[0]['content'] }}\n    {%- else %}\n        {{- 'You are Qwen, created by Alibaba Cloud. You are a helpful assistant.' }}\n    {%- endif %}\n    {{- \"\\n\\n# Tools\\n\\nYou may call one or more functions to assist with the user query.\\n\\nYou are provided with function signatures within <tools></tools> XML tags:\\n<tools>\" }}\n    {%- for tool in tools %}\n        {{- \"\\n\" }}\n        {{- tool | tojson }}\n    {%- endfor %}\n    {{- \"\\n</tools>\\n\\nFor each function call, return a json object with function name and arguments within <tool_call></tool_call> XML tags:\\n<tool_call>\\n{\\\"name\\\": <function-name>, \\\"arguments\\\": <args-json-object>}\\n</tool_call><|im_end|>\\n\" }}\n{%- else %}\n    {%- if messages[0]['role'] == 'system' %}\n        {{- '<|im_start|>system\\n' + messages[0]['content'] + '<|im_end|>\\n' }}\n    {%- else %}\n        {{- '<|im_start|>system\\nYou are Qwen, created by Alibaba Cloud. You are a helpful assistant.<|im_end|>\\n' }}\n    {%- endif %}\n{%- endif %}\n{%- for message in messages %}\n    {%- if (message.role == \"user\") or (message.role == \"system\" and not loop.first) or (message.role == \"assistant\" and not message.tool_calls) %}\n        {{- '<|im_start|>' + message.role + '\\n' + message.content + '<|im_end|>' + '\\n' }}\n    {%- elif message.role == \"assistant\" %}\n        {{- '<|im_start|>' + message.role }}\n        {%- if message.content %}\n            {{- '\\n' + message.content }}\n        {%- endif %}\n        {%- for tool_call in message.tool_calls %}\n            {%- if tool_call.function is defined %}\n                {%- set tool_call = tool_call.function %}\n            {%- endif %}\n            {{- '\\n<tool_call>\\n{\"name\": \"' }}\n            {{- tool_call.name }}\n            {{- '\", \"arguments\": ' }}\n            {{- tool_call.arguments | tojson }}\n            {{- '}\\n</tool_call>' }}\n        {%- endfor %}\n        {{- '<|im_end|>\\n' }}\n    {%- elif message.role == \"tool\" %}\n        {%- if (loop.index0 == 0) or (messages[loop.index0 - 1].role != \"tool\") %}\n            {{- '<|im_start|>user' }}\n        {%- endif %}\n        {{- '\\n<tool_response>\\n' }}\n        {{- message.content }}\n        {{- '\\n</tool_response>' }}\n        {%- if loop.last or (messages[loop.index0 + 1].role != \"tool\") %}\n            {{- '<|im_end|>\\n' }}\n        {%- endif %}\n    {%- endif %}\n{%- endfor %}\n{%- if add_generation_prompt %}\n    {{- '<|im_start|>assistant\\n' }}\n{%- endif %}\n"


class Qwen2dot5ChatTemplate(ChatTemplate):
    # available_roles = ["system", "user", "assistant"]
    template_str = qwen2dot5_it_template

    def get_role_start(self, role_name):
        if role_name in ["system", "user", "assistant"]:
            return f"<|im_start|>{role_name}\n"
        else:
            raise UnsupportedRoleException(role_name, self)

    def get_role_end(self, role_name=None):  # noqa ARG002
        return "<|im_end|>\n"


CHAT_TEMPLATE_CACHE[qwen2dot5_template] = Qwen2dot5ChatTemplate
CHAT_TEMPLATE_CACHE[qwen2dot5_it_template] = Qwen2dot5ChatTemplate


# --------------------------------------------------
# @@@@ Qwen3 @@@@
# --------------------------------------------------
# From https://huggingface.co/Qwen/Qwen3-8B/blob/main/tokenizer_config.json#L230
qwen3_it_template = "{%- if tools %}\n    {{- '<|im_start|>system\\n' }}\n    {%- if messages[0].role == 'system' %}\n        {{- messages[0].content + '\\n\\n' }}\n    {%- endif %}\n    {{- \"# Tools\\n\\nYou may call one or more functions to assist with the user query.\\n\\nYou are provided with function signatures within <tools></tools> XML tags:\\n<tools>\" }}\n    {%- for tool in tools %}\n        {{- \"\\n\" }}\n        {{- tool | tojson }}\n    {%- endfor %}\n    {{- \"\\n</tools>\\n\\nFor each function call, return a json object with function name and arguments within <tool_call></tool_call> XML tags:\\n<tool_call>\\n{\\\"name\\\": <function-name>, \\\"arguments\\\": <args-json-object>}\\n</tool_call><|im_end|>\\n\" }}\n{%- else %}\n    {%- if messages[0].role == 'system' %}\n        {{- '<|im_start|>system\\n' + messages[0].content + '<|im_end|>\\n' }}\n    {%- endif %}\n{%- endif %}\n{%- set ns = namespace(multi_step_tool=true, last_query_index=messages|length - 1) %}\n{%- for message in messages[::-1] %}\n    {%- set index = (messages|length - 1) - loop.index0 %}\n    {%- if ns.multi_step_tool and message.role == \"user\" and message.content is string and not(message.content.startswith('<tool_response>') and message.content.endswith('</tool_response>')) %}\n        {%- set ns.multi_step_tool = false %}\n        {%- set ns.last_query_index = index %}\n    {%- endif %}\n{%- endfor %}\n{%- for message in messages %}\n    {%- if message.content is string %}\n        {%- set content = message.content %}\n    {%- else %}\n        {%- set content = '' %}\n    {%- endif %}\n    {%- if (message.role == \"user\") or (message.role == \"system\" and not loop.first) %}\n        {{- '<|im_start|>' + message.role + '\\n' + content + '<|im_end|>' + '\\n' }}\n    {%- elif message.role == \"assistant\" %}\n        {%- set reasoning_content = '' %}\n        {%- if message.reasoning_content is string %}\n            {%- set reasoning_content = message.reasoning_content %}\n        {%- else %}\n            {%- if '</think>' in content %}\n                {%- set reasoning_content = content.split('</think>')[0].rstrip('\\n').split('<think>')[-1].lstrip('\\n') %}\n                {%- set content = content.split('</think>')[-1].lstrip('\\n') %}\n            {%- endif %}\n        {%- endif %}\n        {%- if loop.index0 > ns.last_query_index %}\n            {%- if loop.last or (not loop.last and reasoning_content) %}\n                {{- '<|im_start|>' + message.role + '\\n<think>\\n' + reasoning_content.strip('\\n') + '\\n</think>\\n\\n' + content.lstrip('\\n') }}\n            {%- else %}\n                {{- '<|im_start|>' + message.role + '\\n' + content }}\n            {%- endif %}\n        {%- else %}\n            {{- '<|im_start|>' + message.role + '\\n' + content }}\n        {%- endif %}\n        {%- if message.tool_calls %}\n            {%- for tool_call in message.tool_calls %}\n                {%- if (loop.first and content) or (not loop.first) %}\n                    {{- '\\n' }}\n                {%- endif %}\n                {%- if tool_call.function %}\n                    {%- set tool_call = tool_call.function %}\n                {%- endif %}\n                {{- '<tool_call>\\n{\"name\": \"' }}\n                {{- tool_call.name }}\n                {{- '\", \"arguments\": ' }}\n                {%- if tool_call.arguments is string %}\n                    {{- tool_call.arguments }}\n                {%- else %}\n                    {{- tool_call.arguments | tojson }}\n                {%- endif %}\n                {{- '}\\n</tool_call>' }}\n            {%- endfor %}\n        {%- endif %}\n        {{- '<|im_end|>\\n' }}\n    {%- elif message.role == \"tool\" %}\n        {%- if loop.first or (messages[loop.index0 - 1].role != \"tool\") %}\n            {{- '<|im_start|>user' }}\n        {%- endif %}\n        {{- '\\n<tool_response>\\n' }}\n        {{- content }}\n        {{- '\\n</tool_response>' }}\n        {%- if loop.last or (messages[loop.index0 + 1].role != \"tool\") %}\n            {{- '<|im_end|>\\n' }}\n        {%- endif %}\n    {%- endif %}\n{%- endfor %}\n{%- if add_generation_prompt %}\n    {{- '<|im_start|>assistant\\n' }}\n    {%- if enable_thinking is defined and enable_thinking is false %}\n        {{- '<think>\\n\\n</think>\\n\\n' }}\n    {%- endif %}\n{%- endif %}"


class Qwen3ChatTemplate(ChatTemplate):
    # available_roles = ["system", "user", "assistant"]
    template_str = qwen2dot5_it_template

    def get_role_start(self, role_name):
        if role_name in ["system", "user", "assistant"]:
            return f"<|im_start|>{role_name}\n"
        else:
            raise UnsupportedRoleException(role_name, self)

    def get_role_end(self, role_name=None):  # noqa ARG002
        return "<|im_end|>\n"


CHAT_TEMPLATE_CACHE[qwen3_it_template] = Qwen3ChatTemplate


# --------------------------------------------------
# @@@@ Llama3.2 @@@@
# --------------------------------------------------
# TODO WIP [HN]: Llama 3.2 has the same issues as Phi-3 mini, where trim behavior is defined in the template.
# Additionally, they use some crazy jinja trick to dynamically pull the current date (but have a fallback of July 26, 2024).
# There's an additional problem that system roles are ALWAYS defined. Might need a refactor the entire system to handle.

llama3dot2_template = '{{- bos_token }}\n{%- if custom_tools is defined %}\n    {%- set tools = custom_tools %}\n{%- endif %}\n{%- if not tools_in_user_message is defined %}\n    {%- set tools_in_user_message = true %}\n{%- endif %}\n{%- if not date_string is defined %}\n    {%- if strftime_now is defined %}\n        {%- set date_string = strftime_now("%d %b %Y") %}\n    {%- else %}\n        {%- set date_string = "26 Jul 2024" %}\n    {%- endif %}\n{%- endif %}\n{%- if not tools is defined %}\n    {%- set tools = none %}\n{%- endif %}\n\n{#- This block extracts the system message, so we can slot it into the right place. #}\n{%- if messages[0][\'role\'] == \'system\' %}\n    {%- set system_message = messages[0][\'content\']|trim %}\n    {%- set messages = messages[1:] %}\n{%- else %}\n    {%- set system_message = "" %}\n{%- endif %}\n\n{#- System message #}\n{{- "<|start_header_id|>system<|end_header_id|>\\n\\n" }}\n{%- if tools is not none %}\n    {{- "Environment: ipython\\n" }}\n{%- endif %}\n{{- "Cutting Knowledge Date: December 2023\\n" }}\n{{- "Today Date: " + date_string + "\\n\\n" }}\n{%- if tools is not none and not tools_in_user_message %}\n    {{- "You have access to the following functions. To call a function, please respond with JSON for a function call." }}\n    {{- \'Respond in the format {"name": function name, "parameters": dictionary of argument name and its value}.\' }}\n    {{- "Do not use variables.\\n\\n" }}\n    {%- for t in tools %}\n        {{- t | tojson(indent=4) }}\n        {{- "\\n\\n" }}\n    {%- endfor %}\n{%- endif %}\n{{- system_message }}\n{{- "<|eot_id|>" }}\n\n{#- Custom tools are passed in a user message with some extra guidance #}\n{%- if tools_in_user_message and not tools is none %}\n    {#- Extract the first user message so we can plug it in here #}\n    {%- if messages | length != 0 %}\n        {%- set first_user_message = messages[0][\'content\']|trim %}\n        {%- set messages = messages[1:] %}\n    {%- else %}\n        {{- raise_exception("Cannot put tools in the first user message when there\'s no first user message!") }}\n{%- endif %}\n    {{- \'<|start_header_id|>user<|end_header_id|>\\n\\n\' -}}\n    {{- "Given the following functions, please respond with a JSON for a function call " }}\n    {{- "with its proper arguments that best answers the given prompt.\\n\\n" }}\n    {{- \'Respond in the format {"name": function name, "parameters": dictionary of argument name and its value}.\' }}\n    {{- "Do not use variables.\\n\\n" }}\n    {%- for t in tools %}\n        {{- t | tojson(indent=4) }}\n        {{- "\\n\\n" }}\n    {%- endfor %}\n    {{- first_user_message + "<|eot_id|>"}}\n{%- endif %}\n\n{%- for message in messages %}\n    {%- if not (message.role == \'ipython\' or message.role == \'tool\' or \'tool_calls\' in message) %}\n        {{- \'<|start_header_id|>\' + message[\'role\'] + \'<|end_header_id|>\\n\\n\'+ message[\'content\'] | trim + \'<|eot_id|>\' }}\n    {%- elif \'tool_calls\' in message %}\n        {%- if not message.tool_calls|length == 1 %}\n            {{- raise_exception("This model only supports single tool-calls at once!") }}\n        {%- endif %}\n        {%- set tool_call = message.tool_calls[0].function %}\n        {{- \'<|start_header_id|>assistant<|end_header_id|>\\n\\n\' -}}\n        {{- \'{"name": "\' + tool_call.name + \'", \' }}\n        {{- \'"parameters": \' }}\n        {{- tool_call.arguments | tojson }}\n        {{- "}" }}\n        {{- "<|eot_id|>" }}\n    {%- elif message.role == "tool" or message.role == "ipython" %}\n        {{- "<|start_header_id|>ipython<|end_header_id|>\\n\\n" }}\n        {%- if message.content is mapping or message.content is iterable %}\n            {{- message.content | tojson }}\n        {%- else %}\n            {{- message.content }}\n        {%- endif %}\n        {{- "<|eot_id|>" }}\n    {%- endif %}\n{%- endfor %}\n{%- if add_generation_prompt %}\n    {{- \'<|start_header_id|>assistant<|end_header_id|>\\n\\n\' }}\n{%- endif %}\n'


class Llama3dot2ChatTemplate(ChatTemplate):
    # available_roles = ["user", "assistant"]
    template_str = llama3dot2_template

    def get_role_start(self, role_name):
        if role_name == "system":
            # TODO: Figure out how Jinja detects datetime and replace with strftime() if needed - currently using fallback
            return "<|start_header_id|>system<|end_header_id|>\n\nCutting Knowledge Date: December 2023\nToday Date: 26 Jul 2024"
        elif role_name == "user":
            return "<|start_header_id|>user<|end_header_id|>\n\n"
        elif role_name == "assistant":
            return "<|start_header_id|>assistant<|end_header_id|>\n\n"
        else:
            raise UnsupportedRoleException(role_name, self)

    def get_role_end(self, role_name=None):  # noqa ARG002
        return "<|eot_id|>"


CHAT_TEMPLATE_CACHE[llama3dot2_template] = Llama3dot2ChatTemplate



"""Renderer is responsible for displaying output.

Our main focus is on jupyter notebooks and later terminal.
"""
# NOTE(nopdive): Testing this notebook related components is tricky. Should figure this out at some point.

import asyncio
import logging
import traceback
from asyncio import Queue
from functools import lru_cache, partial
from importlib.util import find_spec
from typing import TYPE_CHECKING, Optional
from warnings import warn
from weakref import ReferenceType, WeakKeyDictionary, WeakValueDictionary, finalize, ref

from .._topics import DEFAULT_TOPIC, VISUAL_TOPIC
from .._utils import log_cleanup
from ..trace import TraceHandler
from ..visual import (
    ClientReadyMessage,
    GuidanceMessage,
    ResetDisplayMessage,
    TraceMessage,
)
from . import MetricMessage
from ._environment import Environment
from ._jupyter import ipy_handle_event_once
from ._message import (
    ClientReadyAckMessage,
    ExecutionCompletedMessage,
    ExecutionStartedMessage,
    OutputRequestAckMessage,
    OutputRequestMessage,
    deserialize_message,
    serialize_message,
)

try:
    from IPython.display import display

    ipython_imported = True
except ImportError:
    ipython_imported = False
    display = None


stitch_installed = find_spec("stitch") is not None

if TYPE_CHECKING:
    from stitch import StitchWidget

logger = logging.getLogger(__name__)


# Uncomment the following lines to enable file logging
# import datetime
#
# log_filename = f"widget_debug_{datetime.datetime.now().strftime('%H%M%S')}.log"
# file_handler = logging.FileHandler(log_filename)
# file_handler.setLevel(logging.DEBUG)
# formatter = logging.Formatter("%(asctime)s.%(msecs)03d - %(name)s - %(levelname)s - %(message)s", datefmt="%H:%M:%S")
# file_handler.setFormatter(formatter)
# logger.addHandler(file_handler)
# logger.setLevel(logging.DEBUG)


class Renderer:
    """Renders guidance model to a visual medium."""

    def __init__(self):
        """Initializes."""

    def update(self, message: GuidanceMessage, topic: str = DEFAULT_TOPIC) -> None:
        """Updates renderer with incoming message.

        Args:
            message: Incoming message.
            topic: Topic to update.
        """
        raise NotImplementedError("Update not implemented.")


@lru_cache(maxsize=1)
def _get_src_doc_template() -> str:
    """Returns the source document template for the stitch widget."""
    import importlib.resources as resources

    import guidance

    path = resources.files(guidance) / "resources" / "graphpaper-inline.html"
    with path.open("r", encoding="utf-8") as f:
        return f.read()


def _create_stitch_widget() -> "StitchWidget":
    from stitch import StitchWidget

    w = StitchWidget()
    w.initial_width = "100%"
    w.initial_height = "auto"
    w.srcdoc = _get_src_doc_template()
    finalize(w, log_cleanup, f"stitch({id(w)})")

    return w


def _put_nowait_queue(queue: Queue, val: object) -> None:
    from ..registry import get_bg_async

    get_bg_async().call_soon_threadsafe(queue.put_nowait, val)


def _cleanup(recv_queue: Optional[Queue], send_queue: Optional[Queue], log_msg: str, exchange_cb) -> None:
    from ..registry import get_exchange

    log_cleanup(log_msg)
    if send_queue is not None:
        _put_nowait_queue(send_queue, (None, None))
    if recv_queue is not None:
        _put_nowait_queue(recv_queue, (None, None))

    get_exchange().unsubscribe(exchange_cb)


async def _create_queue() -> Queue:
    # This will run in the visual thread.
    return Queue()


def _on_stitch_clientmsg(recv_queue_weakref: ReferenceType["Queue"], identifier: str, change: dict) -> None:
    # NOTE(nopdive): Widget callbacks do not print to stdout/stderr nor module log.
    queue = recv_queue_weakref()
    if queue is not None:
        _put_nowait_queue(queue, (change["new"], identifier))


def _on_cell_completion(renderer_weakref: ReferenceType["JupyterWidgetRenderer"], info) -> None:
    logger.debug("CELL:executed")

    try:
        renderer = renderer_weakref()
        if renderer is None:
            return

        widget = renderer.widgets[renderer.last_widget_key]
        last_trace_id = renderer.widget_to_trace_id[widget]
        message = ExecutionCompletedMessage(
            last_trace_id=last_trace_id,
            is_err=info.error_in_exec is not None,
        )
        renderer.update(message)
    except Exception:
        logger.error(f"CELL_COMPLETE:{traceback.format_exc()}")


async def _handle_recv_messages(
    renderer_weakref: ReferenceType["JupyterWidgetRenderer"], queue_weakref: ReferenceType["Queue"]
) -> None:
    logger.debug("RECV:init")
    from ..registry import get_exchange

    while True:
        try:
            queue = queue_weakref()
            if queue is None:
                break
            value, identifier = await queue.get()
            # logger.debug(f"RECV:raw:{value}")

            if value is None:
                logger.debug("RECV:closing")
                break

            message = deserialize_message(value)
            # logger.debug(f"RECV:msg:{message}")

            renderer = renderer_weakref()
            if renderer is None:
                logger.debug("RECV:renderer early clean")
                break

            if isinstance(message, OutputRequestMessage):
                # Add identifier to message
                message.identifier = identifier
                logger.debug(f"RECV:outputrequest({identifier})")

            get_exchange().publish(message, VISUAL_TOPIC)

            if isinstance(message, ClientReadyMessage):
                logger.debug(f"RECV:clientready({identifier})")
                get_exchange().publish(ClientReadyAckMessage(), VISUAL_TOPIC)

            renderer.recv_queue.task_done()
        except Exception as _:
            logger.error(f"RECV:err:{traceback.format_exc()}")


async def _handle_send_messages(
    renderer_weakref: ReferenceType["JupyterWidgetRenderer"], queue_weakref: ReferenceType["Queue"]
) -> None:
    logger.debug("SEND:init")
    # NOTE(nopdive): Waiting on client cb does not work, client messages received on cell completion.
    #                Currently, we do a replay of messages on completion for client if client
    #                first receives non-zero message identifier.

    # What if we only used 1% of our brain?
    await asyncio.sleep(200 / 1000.0)
    logger.debug("SEND:ready")

    while True:
        try:
            queue = queue_weakref()
            if queue is None:
                break
            message, identifier = await queue.get()
            # logger.debug(f"SEND:msg:{message}")

            if message is None:
                logger.debug("SEND:closing")
                break

            message_json = serialize_message(message)
            # logger.debug(f"SEND:json:{message_json}")

            renderer = renderer_weakref()
            if renderer is None:
                break
            if isinstance(renderer, JupyterWidgetRenderer):
                widget_key = renderer.last_widget_key if identifier is None else identifier
                if renderer.widgets.get(widget_key, None) is not None:
                    # NOTE(nopdive): This at random times, appears to fire two changes instead of one change event.
                    renderer.widgets[widget_key].kernelmsg = message_json
            else:
                logger.debug("SEND:jupyter:send but no widget")
            renderer.send_queue.task_done()
        except Exception as _:
            logger.error(f"SEND:err:{traceback.format_exc()}")


def _trace_path_to_messages(trace_id: int) -> list["TraceMessage"]:
    """Convert trace path from root to given trace_id into TraceMessage objects.

    Args:
        trace_id: The trace ID to get the path for.

    Returns:
        List of TraceMessage objects representing the path from root to the given trace_id.

    """
    from ..registry import get_trace_handler

    trace_handler = get_trace_handler()
    trace_path = trace_handler.id_node_map[trace_id].path()
    messages = []

    for trace_node in trace_path:
        node_trace_id = trace_handler.node_id_map[trace_node]
        parent_trace_id = None
        if trace_node.parent is not None:
            parent_trace_id = trace_handler.node_id_map[trace_node.parent]

        for input_attr in trace_node.input:
            input_message = TraceMessage(trace_id=node_trace_id, parent_trace_id=parent_trace_id, node_attr=input_attr)
            messages.append(input_message)

        for output_attr in trace_node.output:
            output_message = TraceMessage(
                trace_id=node_trace_id, parent_trace_id=parent_trace_id, node_attr=output_attr
            )
            messages.append(output_message)

    return messages


class JupyterWidgetRenderer(Renderer):
    """Jupyter widget renderer that is implemented via stitch package."""

    def __init__(self, trace_handler: TraceHandler) -> None:
        """Initializes.

        Args:
            trace_handler: Trace handler of an engine.
        """
        from ..registry import get_bg_async, get_exchange

        super().__init__()
        self.widget_to_trace_id: WeakKeyDictionary["StitchWidget", int] = WeakKeyDictionary()
        self.widget_messages: WeakKeyDictionary["StitchWidget", list[GuidanceMessage]] = WeakKeyDictionary()
        self.widgets: WeakValueDictionary[str, "StitchWidget"] = WeakValueDictionary()
        self.last_widget_key = ""
        self.last_widget: Optional[ReferenceType["StitchWidget"]] = None
        self.last_trace_id: Optional[int] = None
        self._last_cell_session_id: Optional[str] = None

        self._trace_handler = trace_handler
        self._completed = False
        self._running = False

        # Debug tracking
        self._debug_enabled = False
        self._debug_messages: list[GuidanceMessage] = []

        # Create queue and wait for instantiation
        self.send_queue: Queue = get_bg_async().run_async_coroutine(_create_queue()).result()
        self.recv_queue: Queue = get_bg_async().run_async_coroutine(_create_queue()).result()

        # Start send/recv message loops
        recv_coroutine = _handle_recv_messages(ref(self), ref(self.recv_queue))
        send_coroutine = _handle_send_messages(ref(self), ref(self.send_queue))
        self._recv_task = get_bg_async().run_async_coroutine(get_bg_async().async_task(recv_coroutine)).result()
        self._send_task = get_bg_async().run_async_coroutine(get_bg_async().async_task(send_coroutine)).result()

        # Start recv on exchange
        get_exchange().subscribe(self._on_exchange)
        finalize(self, _cleanup, self.recv_queue, self.send_queue, f"renderer({id(self)})", self._on_exchange)

    def _on_exchange(self, message: GuidanceMessage) -> None:
        # if not isinstance(message, MetricMessage):  # NOTE(nopdive): Metrics spam at fixed intervals.
        #     logger.debug(f"RENDERER_ON_EXCHANGE:{message}")

        if isinstance(message, MetricMessage):
            self.update(message)
        elif isinstance(message, TraceMessage):
            self.update(message)
        elif isinstance(message, OutputRequestMessage):
            self._replay(message)

    def has_divergence(self, message: GuidanceMessage) -> tuple[bool, int]:
        """Checks if message has divergence with current path.

        Args:
            message: Incoming message.

        Returns:
            tuple of (has diverged, shared ancestor index). Index will be -1 if divergence and requires trace replay.
        """
        if not isinstance(message, TraceMessage):
            return False, -1

        # If we diverge from the model path, truncate and reset
        message_trace_node = self._trace_handler[message.trace_id]
        widget_messages = self.widget_messages[self.last_widget()]

        prev_trace_messages = [x for x in widget_messages if isinstance(x, TraceMessage)]
        trace_messages_len = len(prev_trace_messages)
        if trace_messages_len == 0:
            return False, -1
        elif trace_messages_len == 1:
            if isinstance(widget_messages[0], TraceMessage):
                try:
                    last_trace_node = self._trace_handler[widget_messages[0].trace_id]
                    if message_trace_node.parent == last_trace_node:
                        return False, -1
                    else:
                        return True, 0
                except KeyError:
                    # Trace node was garbage collected, treat as divergence
                    return True, 0
            else:
                return False, -1
        else:
            last_trace_message = prev_trace_messages[-1]
            try:
                last_trace_node = self._trace_handler[last_trace_message.trace_id]
            except KeyError:
                # Trace node was garbage collected, treat as divergence
                return True, 0

            message_trace_node_path = message_trace_node.path()
            if last_trace_node not in message_trace_node_path:
                logger.debug(f"DIVERGENCE:curr:{message_trace_node}")
                logger.debug(f"DIVERGENCE:prev:{last_trace_node}")

                # Truncate path that is no longer used by current trace node
                ancestor_idx = -1
                ancestors = set(message_trace_node.ancestors())
                for idx, prev_message in enumerate(widget_messages):
                    if isinstance(prev_message, TraceMessage):
                        try:
                            prev_trace_node = self._trace_handler[prev_message.trace_id]
                            if prev_trace_node in ancestors:
                                ancestor_idx = idx
                        except KeyError:
                            # Trace node was garbage collected, skip it
                            continue
                if ancestor_idx == -1:
                    if message_trace_node.parent == last_trace_node.root():  # pragma: no cover
                        ancestor_idx = 0
                    else:
                        logger.debug("DIVERGENCE:full reset (not in messages)")
                        ancestor_idx = -1

                return True, ancestor_idx
            else:
                return False, -1

    def _replay(self, message: OutputRequestMessage, topic=DEFAULT_TOPIC) -> None:
        out_messages: list[GuidanceMessage] = []
        logger.debug(f"RENDERER:replay:{message}")

        widget = self.widgets.get(message.identifier, None)
        if widget is None:
            logger.debug(f"RENDERER:widget({message.identifier}) already gc")
            return

        widget_messages = self.widget_messages[widget]
        out_messages.append(OutputRequestAckMessage())
        out_messages.append(ResetDisplayMessage())
        out_messages.extend(widget_messages)
        widget_messages.clear()

        for out_message in out_messages:
            if isinstance(out_message, TraceMessage):
                self.widget_to_trace_id[widget] = out_message.trace_id
            widget_messages.append(out_message)
            _put_nowait_queue(self.send_queue, (out_message, widget.model_id))

    def update(self, message: GuidanceMessage, topic=DEFAULT_TOPIC) -> None:
        out_messages: list[GuidanceMessage] = []

        # Short-circuit for metrics
        if isinstance(message, MetricMessage):
            if not self._running:
                return

        # Create new widget if not already running (first display when a cell starts execution)
        if not self._running and isinstance(message, TraceMessage):
            # Execution started
            started_msg = ExecutionStartedMessage()
            out_messages.append(started_msg)
            out_messages.append(MetricMessage(name="status", value="Running"))

            logger.debug("RENDERER:new widget needed")
            _, last_cell_session_id = ipy_handle_event_once(partial(_on_cell_completion, ref(self)), "post_run_cell")
            widget = _create_stitch_widget()
            widget_key = widget.model_id
            widget_messages = []
            if self.last_widget is not None:
                last_widget = self.last_widget()
                if last_widget is not None:
                    widget_messages[:] = self.widget_messages[last_widget][:]
            self.widgets[widget_key] = widget
            self.widget_messages[widget] = widget_messages
            self.widget_to_trace_id[widget] = message.trace_id
            self.last_widget_key = widget_key
            self.last_widget = ref(widget)

            on_clientmsg = partial(_on_stitch_clientmsg, ref(self.recv_queue), widget_key)
            widget.observe(on_clientmsg, names="clientmsg")

            # Redraw
            display(widget)
            logger.debug(f"RENDERER:widget displayed")

            self._last_cell_session_id = last_cell_session_id
            self._running = True
            self._completed = False

            # NOTE(nopdive): Fire off execution immediately to renderer subscribers. Review later.
            _put_nowait_queue(self.recv_queue, (serialize_message(started_msg), self.last_widget_key))

        widget_messages = self.widget_messages[self.last_widget()]

        # This is fired when the cell completes execution
        if isinstance(message, ExecutionCompletedMessage):
            # Execution completed
            logger.debug("RENDERER:execution end")
            self._completed = True
            self._running = False

            # Replay all messages on completion
            out_messages.append(ResetDisplayMessage())
            out_messages[len(out_messages) :] = widget_messages[:]
            widget_messages.clear()

            if message.is_err:
                out_messages.append(MetricMessage(name="status", value="Error"))
            else:
                out_messages.append(MetricMessage(name="status", value="Done"))

        # Check if message has diverged from prev messages
        diverged, shared_ancestor_idx = self.has_divergence(message)
        if diverged:
            logger.debug(f"RENDERER:diverged, shared ancestor idx: {shared_ancestor_idx}")
            out_messages.append(ResetDisplayMessage())
            if shared_ancestor_idx >= 0:
                out_messages[len(out_messages) :] = widget_messages[: shared_ancestor_idx + 1]
                widget_messages.clear()
            else:
                logger.debug("RENDERER:diverged, but no shared ancestor, replay")
                # Reconstruct trace messages and replay from root
                trace_messages = _trace_path_to_messages(message.trace_id)
                out_messages.extend(trace_messages)
                widget_messages.clear()

        # Append current message to outgoing
        out_messages.append(message)

        # Send outgoing messages to client
        widget = self.widgets[self.last_widget_key]
        for out_message in out_messages:
            if isinstance(out_message, TraceMessage):
                self.widget_to_trace_id[widget] = out_message.trace_id
            widget_messages.append(out_message)
            _put_nowait_queue(self.send_queue, (out_message, self.last_widget_key))

            # Track for debug if enabled
            if self._debug_enabled:
                self._debug_messages.append(out_message)

    # TODO(nopdive): I'm not sure how we should debug here with the new changes, Nick to review.
    def enable_debug(self) -> None:
        """Enable debug mode in the widget to capture message history."""

        self._debug_enabled = True
        self._debug_messages = []  # Clear previous messages

    def clear_debug_data(self) -> None:
        """Clear captured debug messages."""
        self._debug_messages = []
        logger.info("Debug messages cleared")

    def get_debug_data(self) -> Optional[str]:
        """Get debug data as a JSON string.

        Returns:
            JSON string containing debug data, or None if no data available
        """
        import datetime
        import json

        if not self._debug_enabled:
            logger.warning("Debug mode not enabled - call enable_debug() first")
            return None

        if not self._debug_messages:
            logger.warning("No debug messages captured yet")
            return None

        def make_json_serializable(obj):
            """Convert objects to JSON serializable format."""
            if isinstance(obj, bytes):
                # Convert bytes to base64 string for JSON serialization
                import base64

                return base64.b64encode(obj).decode("utf-8")
            elif isinstance(obj, dict):
                return {k: make_json_serializable(v) for k, v in obj.items()}
            elif isinstance(obj, list):
                return [make_json_serializable(item) for item in obj]
            elif hasattr(obj, "model_dump"):
                # Pydantic models
                return make_json_serializable(obj.model_dump())
            else:
                return obj

        debug_data = {
            "timestamp": datetime.datetime.now().isoformat(),
            "messageCount": len(self._debug_messages),
            "messages": [
                make_json_serializable({"message_id": msg.message_id, "class_name": msg.class_name, **msg.model_dump()})
                for msg in self._debug_messages
            ],
        }

        return json.dumps(debug_data, indent=2)


class DoNothingRenderer(Renderer):
    """It does nothing."""

    def __init__(self, trace_handler: TraceHandler) -> None:
        """Initializes.

        Args:
            trace_handler: Trace handler of an engine.
        """
        self._trace_handler = trace_handler
        super().__init__()

    def update(self, message: GuidanceMessage, topic: str = DEFAULT_TOPIC) -> None:
        pass


class AutoRenderer(Renderer):
    """Automatically detects which renderer to use based on environment."""

    def __init__(self, trace_handler: TraceHandler):
        """Initializes.

        Args:
            trace_handler: Trace handler of an engine.
        """
        self._env = Environment()

        self._renderer: Renderer
        if self._env.is_notebook():
            if stitch_installed:
                self._renderer = JupyterWidgetRenderer(trace_handler=trace_handler)
            else:
                self._renderer = DoNothingRenderer(trace_handler=trace_handler)
        elif self._env.is_terminal():
            # TODO(nopdive): When IPython events are figured out (cell completion)
            #                hook up terminal interface separate to non-interactive
            #                shell.
            self._renderer = DoNothingRenderer(trace_handler=trace_handler)
        else:  # pragma: no cover
            logger.error("Env detection has failed. This is a bug.")
            warn("Env detection has failed. No renderer will be provided.")
            self._renderer = DoNothingRenderer(trace_handler=trace_handler)

        super().__init__()

    def update(self, message: GuidanceMessage, topic: str = DEFAULT_TOPIC) -> None:
        self._renderer.update(message, topic=topic)

    def renderer_type(self) -> type:
        """Type of renderer that has been selected based on environment.

        Returns:
            Type of selected renderer.
        """
        return type(self._renderer)



"""Jupyter specific utilities."""

import logging
from typing import Any, Callable, Optional
from uuid import uuid4

try:
    from IPython import get_ipython
except ImportError:
    pass

logger = logging.getLogger(__name__)

IPythonCallback = Callable[[Any], None]


def ipy_handle_event_once(cb: IPythonCallback, event_name: str) -> tuple[Optional[IPythonCallback], str]:
    ipy = get_ipython()
    cell_session_id = str(uuid4())

    if ipy is None:
        return None, ""

    def cb_closure(msg):
        cb(info=msg)
        ipy.events.unregister(event_name, cb_closure)

    ipy.events.register(event_name, cb_closure)

    return cb_closure, cell_session_id



import base64
import importlib.resources
import pathlib
import typing

from .._guidance import guidance
from .._utils import bytes_from

# from ..trace._trace import VideoInput
from ..trace._trace import VideoOutput


@guidance
def video(lm, src: typing.Union[str, pathlib.Path, bytes], allow_local: bool = True):
    # TODO(nopdive): Mock for testing. Remove all of this code later.
    bytes_data = bytes_from(src, allow_local=allow_local)
    base64_bytes = base64.b64encode(bytes_data)
    lm += VideoOutput(value=base64_bytes, is_input=True)
    # lm += VideoInput(value=base64_string)
    return lm


@guidance
def gen_video(lm):
    # TODO(nopdive): Mock for testing. Remove all of this code later.
    with importlib.resources.files("guidance").joinpath("resources/sample_video.png").open("rb") as f:
        bytes_data = f.read()
    base64_bytes = base64.b64encode(bytes_data)
    lm += VideoOutput(value=base64_bytes, is_input=False)
    return lm



__version__ = "0.2.4"

import sys
import types

from . import library, models
from ._guidance import guidance

# we expose all the library functions at the top level of the module
from .library import *  # noqa: F403

__all__ = ["guidance", "models", "library", *library.__all__]


# This makes the guidance module callable
class _Guidance(types.ModuleType):
    def __call__(self, f=None, *, stateless=False, cache=None, dedent=True, model=models.Model):
        return guidance(f, stateless=stateless, cache=cache, dedent=dedent, model=model)


sys.modules[__name__].__class__ = _Guidance



import base64
import time
import wave
from abc import ABC, abstractmethod
from copy import deepcopy
from io import BytesIO
from typing import TYPE_CHECKING, Any, ContextManager, Iterator, Literal, Optional, Union, cast

from pydantic import BaseModel, Discriminator, Field, TypeAdapter
from typing_extensions import Annotated, assert_never

from .._ast import (
    ASTNode,
    AudioBlob,
    GenAudio,
    ImageBlob,
    ImageUrl,
    JsonNode,
    LiteralNode,
    RegexNode,
    RoleEnd,
    RoleStart,
    RuleNode,
)
from .._schema import TokenUsage
from .._utils import bytes_from
from ..trace import AudioOutput, ImageOutput, OutputAttr, TextOutput, Token, TokenOutput
from ._base import Interpreter, State

if TYPE_CHECKING:
    import openai
    from openai.types.chat import ChatCompletionChunk


def get_role_start(role: str) -> str:
    # ChatML is as good as anything
    return "<|im_start|>" + role + "\n"


def get_role_end(role: str) -> str:
    # ChatML is as good as anything
    return "\n<|im_end|>\n"


class AssistantAudio(BaseModel):
    id: str
    expires_at: int = Field(exclude=True)
    data: str = Field(exclude=True)
    transcript: str = Field(exclude=True)


class AssistantAudioMessage(BaseModel):
    role: Literal["assistant"]
    audio: AssistantAudio


class TextContent(BaseModel):
    type: Literal["text"]
    text: str


class InputAudio(BaseModel):
    data: str
    format: str


class AudioContent(BaseModel):
    type: Literal["input_audio"]
    input_audio: InputAudio


class ImageUrlContentInner(BaseModel):
    url: str


class ImageUrlContent(BaseModel):
    type: Literal["image_url"]
    image_url: ImageUrlContentInner


Content = Annotated[Union[TextContent, AudioContent, ImageUrlContent], Discriminator("type")]


class ContentMessage(BaseModel):
    role: Literal["system", "user", "assistant"]
    content: list[Content]


Message = Union[ContentMessage, AssistantAudioMessage]


class OpenAIState(State):
    def __init__(self) -> None:
        super().__init__()
        self.messages: list[Message] = []
        self.content: list[Content] = []
        self.audio: Optional[AssistantAudio] = None

    def apply_text(self, text: str) -> None:
        if len(self.content) > 0 and isinstance(self.content[-1], TextContent):
            self.content[-1].text += text
        else:
            self.content.append(TextContent(type="text", text=text))

    def get_active_message(self) -> Optional[Message]:
        if self.active_role is None:
            return None
        if self.active_role not in ["system", "user", "assistant"]:
            raise ValueError(f"Invalid active role: {self.active_role}")
        active_role = cast(Literal["system", "user", "assistant"], self.active_role)

        if self.content and self.audio:
            raise ValueError("Cannot have both content and audio")
        if self.audio:
            if active_role != "assistant":
                raise ValueError("Audio messages can only be sent by the assistant")
            return AssistantAudioMessage(
                role=active_role,
                audio=self.audio,
            )
        elif self.content:
            return ContentMessage(
                role=active_role,
                content=self.content,
            )
        else:
            return None

    def __str__(self) -> str:
        messages = self.messages
        active_message = self.get_active_message()
        if active_message is not None:
            messages = messages + [active_message]
        s = ""
        for i, message in enumerate(messages):
            s += get_role_start(message.role)
            if isinstance(message, AssistantAudioMessage):
                s += "[AUDIO]"
            elif isinstance(message, ContentMessage):
                for content in message.content:
                    if isinstance(content, TextContent):
                        s += content.text
                    elif isinstance(content, ImageUrlContent):
                        s += "[IMAGE]"  # Arbitrary stringification
                    elif isinstance(content, AudioContent):
                        s += "[AUDIO]"  # transcript?
                    else:
                        if TYPE_CHECKING:
                            assert_never(content)
                        raise TypeError(f"Unknown content type: {content}")
            else:
                if TYPE_CHECKING:
                    assert_never(message)
                raise TypeError(f"Unknown message type: {message}")
            if active_message is None or i != len(messages) - 1:
                # For the sake of consistency, don't add role end for the active message
                s += get_role_end(message.role)
        return s


class BaseOpenAIClientWrapper(ABC):
    @abstractmethod
    def streaming_chat_completions(
        self,
        model: str,
        messages: list[dict[str, Any]],
        logprobs: bool,
        **kwargs,
    ) -> ContextManager[Iterator["ChatCompletionChunk"]]:
        """Streaming chat completions."""
        raise NotImplementedError("This method should be implemented by subclasses.")


class OpenAIClientWrapper(BaseOpenAIClientWrapper):
    def __init__(self, client: "openai.OpenAI"):
        self.client = client

    def streaming_chat_completions(
        self,
        model: str,
        messages: list[dict[str, Any]],
        logprobs: bool,
        **kwargs,
    ) -> ContextManager[Iterator["ChatCompletionChunk"]]:
        """Streaming chat completions."""

        return self.client.chat.completions.create(
            model=model,
            messages=messages,
            logprobs=logprobs,
            stream=True,
            stream_options={"include_usage": True},
            **kwargs,
        )


class BaseOpenAIInterpreter(Interpreter[OpenAIState]):
    """Base class for interacting with OpenAI models."""

    logprobs: bool = True
    # TODO: have top-k be passed programmatically and only if echo=True
    top_k: Optional[int] = 5

    def __init__(self, model: str, client: BaseOpenAIClientWrapper, **kwargs):
        super().__init__(state=OpenAIState())
        self.model = model
        self.client = client

    def run(self, node: ASTNode, **kwargs) -> Iterator[OutputAttr]:
        if not isinstance(node, RoleStart) and self.state.active_role is None:
            raise ValueError("OpenAI models require an active role (e.g. use `with assistant(): ...`)")
        return super().run(node, **kwargs)

    def role_start(self, node: RoleStart, **kwargs) -> Iterator[OutputAttr]:
        if node.role not in ["system", "user", "assistant"]:
            raise ValueError(f"OpenAI models only support roles 'system', 'user', and 'assistant', got {node.role}")
        self.state.active_role = cast(Literal["system", "user", "assistant"], node.role)
        # TODO: drop this and yield nothing. We need to add this for now as a workaround for the
        # fact that current vis code assumes that there is actually a role start message
        yield TextOutput(value=get_role_start(node.role), is_input=True)

    def role_end(self, node: RoleEnd, **kwargs) -> Iterator[OutputAttr]:
        active_message = self.state.get_active_message()
        if active_message is not None:
            self.state.messages.append(active_message)
        self.state.audio = None
        self.state.content = []
        self.state.active_role = None
        yield from ()

    def text(self, node: LiteralNode, **kwargs) -> Iterator[OutputAttr]:
        self.state.apply_text(node.value)
        yield TextOutput(value=node.value, is_input=True)

    def _run(self, **kwargs) -> Iterator[OutputAttr]:
        if self.state.active_role is None:
            # Should never happen?
            raise ValueError("OpenAI models require chat blocks (e.g. use `with assistant(): ...`)")
        if self.state.active_role != "assistant":
            raise ValueError(
                "OpenAI models can only generate as the assistant (i.e. inside of `with assistant(): ...`)"
            )
        if self.state.content:
            raise ValueError(
                f"OpenAI models do not support pre-filled assistant messages: got data {self.state.content}."
            )

        sampling_params = kwargs.pop("sampling_params", None)
        if sampling_params:
            # only process kwargs that are supported by the OpenAI API
            if "top_p" not in kwargs:
                kwargs["top_p"] = sampling_params.get("top_p", None)

            if sampling_params.get("top_k", None) is not None:
                raise ValueError("OpenAI models do not support top_k sampling.")

            if sampling_params.get("min_p", None) is not None:
                raise ValueError("OpenAI models do not support min_p sampling.")

            if sampling_params.get("repetition_penalty", None) is not None:
                raise ValueError("OpenAI models do not support repetition_penalty sampling.")

        with self.client.streaming_chat_completions(
            model=self.model,
            messages=cast(list[dict[str, Any]], TypeAdapter(list[Message]).dump_python(self.state.messages)),
            logprobs=self.logprobs,
            top_logprobs=self.top_k if self.logprobs else None,
            **kwargs,
        ) as chunks:
            yield from self._handle_stream(chunks)

    def _handle_stream(self, chunks: Iterator["ChatCompletionChunk"]) -> Iterator[OutputAttr]:
        _t0 = time.time()
        t0 = _t0
        audio: Optional[AssistantAudio] = None
        # We made another call to the OpenAI API, so we count it as a round trip.
        usage = TokenUsage(round_trips=1)
        for chunk in chunks:
            t1 = time.time()
            latency_ms = (t1 - t0) * 1000
            t0 = t1

            # NOTE: use getattr here as litellm does not return usage
            if getattr(chunk, "usage", None) is not None:
                # Update token usage
                usage.input_tokens += chunk.usage.prompt_tokens
                # Estimate forward passes as number of completion tokens
                usage.forward_passes += chunk.usage.completion_tokens
                if getattr(chunk.usage, "prompt_tokens_details", None) is not None:
                    if chunk.usage.prompt_tokens_details.cached_tokens is not None:
                        usage.cached_input_tokens += chunk.usage.prompt_tokens_details.cached_tokens
            if chunk.choices is None or len(chunk.choices) == 0:
                # Azure seems to return empty choices sometimes (on first chunk?)
                # OpenAI seems to return None choices sometimes (after giving usage?) (for audio only?)
                continue
            choice = chunk.choices[0]
            delta = choice.delta
            if delta.content is not None:
                assert audio is None
                content = delta.content
                if len(content) == 0:
                    continue
                self.state.apply_text(content)
                # Rather paranoid check, as we have a few slightly different
                # apis which are "almost" openai compatible...
                if (
                    hasattr(choice, "logprobs")
                    and choice.logprobs is not None
                    and hasattr(choice.logprobs, "content")
                    and choice.logprobs.content is not None
                    and len(choice.logprobs.content) > 0
                ):
                    tokens = choice.logprobs.content
                    for token in tokens:
                        yield TokenOutput(
                            value=content if len(tokens) == 1 else token.token,
                            # amortized latency
                            latency_ms=latency_ms / len(tokens),
                            token=Token(
                                token=token.token,
                                bytes=b"" if token.bytes is None else base64.b64encode(bytes(token.bytes)),
                                prob=2.718**token.logprob,
                            ),
                            top_k=[
                                Token(
                                    token=tok.token,
                                    bytes=b"" if tok.bytes is None else base64.b64encode(bytes(tok.bytes)),
                                    prob=2.718**tok.logprob,
                                )
                                for tok in token.top_logprobs
                            ],
                            is_generated=True,
                        )
                else:
                    yield TextOutput(value=delta.content, is_generated=True, latency_ms=latency_ms)
            elif (delta_audio := cast(Optional[dict], getattr(delta, "audio", None))) is not None:
                transcript_chunk: Optional[str] = None
                if audio is None:
                    assert delta_audio.get("id") is not None
                    audio = AssistantAudio(
                        id=delta_audio["id"],
                        expires_at=delta_audio.get("expires_at", 0),  # ?
                        transcript=delta_audio.get("transcript", ""),
                        data=delta_audio.get("data", ""),
                    )
                    transcript_chunk = delta_audio.get("transcript")
                else:
                    assert delta_audio.get("id") is None or delta_audio["id"] == audio.id
                    if delta_audio.get("data") is not None:
                        audio.data += delta_audio["data"]
                    if delta_audio.get("transcript") is not None:
                        audio.transcript += delta_audio["transcript"]
                        transcript_chunk = delta_audio["transcript"]
                    if delta_audio.get("expires_at") is not None:
                        assert audio.expires_at == 0
                        audio.expires_at = delta_audio["expires_at"]
                if transcript_chunk is not None:
                    # Why not give the users some transcript? :)
                    yield TextOutput(
                        value=delta_audio["transcript"],
                        is_generated=True,
                        latency_ms=latency_ms,
                    )
            elif delta.function_call is not None:
                raise NotImplementedError("Function calling not yet supported for OpenAI")
            elif delta.tool_calls is not None and len(delta.tool_calls) > 0:
                raise NotImplementedError("Tool calling not yet supported for OpenAI")
            # there are cases where vllm does not return refusal field in delta, using None as default value here
            elif getattr(delta, "refusal", None) is not None:
                raise ValueError(f"OpenAI refused the request: {delta.refusal}")

            if choice.finish_reason is not None and choice.finish_reason != "stop":
                # TODO: handle "bad" finish reasons
                pass

        if audio is not None:
            assert self.state.audio is None
            self.state.audio = audio
            # Create an in-memory WAV file
            wav_buffer = BytesIO()
            with wave.open(wav_buffer, "wb") as wav_file:
                wav_file.setnchannels(1)
                wav_file.setsampwidth(2)  # PCM16 = 2 bytes per sample
                wav_file.setframerate(22050)  # A guess
                wav_file.writeframes(base64.b64decode(audio.data))

            # Get WAV bytes
            wav_bytes = wav_buffer.getvalue()
            yield AudioOutput(value=base64.b64encode(wav_bytes), is_input=False)

        usage.total_latency_ms += (time.time() - _t0) * 1000
        self.state.add_usage(usage)

    def __deepcopy__(self, memo):
        """Custom deepcopy to ensure client is not copied."""
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k == "client":
                # Don't copy the client
                setattr(result, k, v)
            else:
                setattr(result, k, deepcopy(v, memo))
        return result


class OpenAIRuleMixin(BaseOpenAIInterpreter):
    def rule(self, node: RuleNode, **kwargs) -> Iterator[OutputAttr]:
        if node.stop:
            raise ValueError("Stop condition not yet supported for OpenAI")
        if node.suffix:
            raise ValueError("Suffix not yet supported for OpenAI")
        if node.stop_capture:
            raise ValueError("Save stop text not yet supported for OpenAI")

        kwargs = kwargs.copy()
        if node.temperature:
            kwargs["temperature"] = node.temperature
        if node.max_tokens:
            kwargs["max_tokens"] = node.max_tokens

        chunks = self.run(node.value, **kwargs)
        if node.capture:
            buffered_text = ""
            for chunk in chunks:
                # TODO: this isinstance check is pretty darn fragile.
                # ~there must be a better way~
                if isinstance(chunk, TextOutput):
                    buffered_text += chunk.value
                yield chunk
            yield self.state.apply_capture(
                name=node.capture,
                value=buffered_text,
                log_prob=1,  # TODO
                is_append=node.list_append,
            )
        else:
            yield from chunks


class OpenAIRegexMixin(BaseOpenAIInterpreter):
    def regex(self, node: RegexNode, **kwargs) -> Iterator[OutputAttr]:
        if node.regex is not None:
            raise ValueError("Regex not yet supported for OpenAI")
        # We're in unconstrained mode now.
        return self._run(**kwargs)


class OpenAIJSONMixin(BaseOpenAIInterpreter):
    def json(self, node: JsonNode, **kwargs) -> Iterator[OutputAttr]:
        if node.schema is None:
            response_format = {"type": "json_object"}
        else:
            response_format = {
                "type": "json_schema",
                "json_schema": {
                    "name": "json_schema",  # TODO?
                    "schema": node.schema,
                    "strict": True,
                },
            }
        return self._run(
            response_format=response_format,
            **kwargs,
        )


class OpenAIImageMixin(BaseOpenAIInterpreter):
    def image_blob(self, node: ImageBlob, **kwargs) -> Iterator[OutputAttr]:
        try:
            import PIL.Image
        except ImportError as ie:
            raise Exception(
                "Please install the Pillow package `pip install Pillow` in order to use images with OpenAI!"
            ) from ie

        image_bytes = base64.b64decode(node.data)
        with PIL.Image.open(BytesIO(image_bytes)) as pil_image:
            # Use PIL to infer file format
            # TODO: just store format on ImageOutput type
            format = pil_image.format
            if format is None:
                raise ValueError("Cannot upload image with unknown format")

        mime_type = f"image/{format.lower()}"
        self.state.content.append(
            ImageUrlContent(
                type="image_url",
                image_url=ImageUrlContentInner(url=f"data:{mime_type};base64,{node.data.decode()}"),
            )
        )
        yield ImageOutput(value=node.data, is_input=True)

    def image_url(self, node: ImageUrl, **kwargs) -> Iterator[OutputAttr]:
        self.state.content.append(ImageUrlContent(type="image_url", image_url=ImageUrlContentInner(url=node.url)))
        image_bytes = bytes_from(node.url, allow_local=False)
        yield ImageOutput(value=base64.b64encode(image_bytes), is_input=True)


class OpenAIAudioMixin(BaseOpenAIInterpreter):
    # Audio models don't support logprobs
    logprobs: bool = False

    def audio_blob(self, node: AudioBlob, **kwargs) -> Iterator[OutputAttr]:
        format = "wav"  # TODO: infer from node
        self.state.content.append(
            AudioContent(
                type="input_audio",
                input_audio=InputAudio(
                    data=node.data.decode("utf-8"),  # Base64 encoded string
                    format=format,
                ),
            )
        )
        yield AudioOutput(value=node.data, format=format, is_input=True)

    def gen_audio(self, node: GenAudio, **kwargs) -> Iterator[OutputAttr]:
        yield from self._run(
            modalities=["text", "audio"],  # Has to be both?
            audio={
                "voice": node.kwargs.get("voice", "alloy"),
                "format": "pcm16",  # Has to be pcm16 for streaming
            },
        )



import atexit
import logging
import operator
import os
import sys
from itertools import takewhile
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Union

import numpy as np

from .._schema import SamplingParams
from .._utils import normalize_notebook_stdout_stderr
from ..chat import ChatTemplate
from ._base import Model
from ._engine import Engine, EngineInterpreter, LogitsOutput, Tokenizer

try:
    import llama_cpp

    is_llama_cpp = True
except ModuleNotFoundError:
    is_llama_cpp = False
else:
    import llguidance.llamacpp

if TYPE_CHECKING:
    from llama_cpp.llama import Llama

logger = logging.getLogger(__name__)

shutdown_none = True


def set_shutdown_flag():
    global shutdown_none
    # python can set anything to None at shutdown, so use None
    shutdown_none = None


atexit.register(set_shutdown_flag)


class _LlamaBatchContext:
    def __init__(self, n_batch, n_ctx):
        self._llama_batch_free = llama_cpp.llama_batch_free
        self.batch = llama_cpp.llama_batch_init(n_batch, 0, n_ctx)
        if self.batch is None:
            raise Exception("call to llama_cpp.llama_batch_init returned NULL.")

    def __del__(self):
        if shutdown_none is not None:
            llama_batch_free = getattr(self, "_llama_batch_free", None)
            batch = getattr(self, "batch", None)
            if batch is not None and llama_batch_free is not None:
                self._llama_batch_free = None
                self.batch = None
                llama_batch_free(batch)


class LlamaCppTokenizer(Tokenizer):
    def __init__(self, model_obj: "Llama", chat_template: Union[str, ChatTemplate, None] = None):
        self._model_obj = model_obj

        vocab = llama_cpp.llama_model_get_vocab(model_obj.model)
        if vocab is None:
            raise Exception("call to llama_cpp.llama_model_get_vocab returned NULL.")
        ll_tokenizer = llguidance.llamacpp.lltokenizer_from_vocab(vocab)

        # Chat Template logic
        if chat_template is None:
            if hasattr(self._model_obj, "metadata") and "tokenizer.chat_template" in self._model_obj.metadata:
                chat_template = self._model_obj.metadata["tokenizer.chat_template"]

        super().__init__(ll_tokenizer=ll_tokenizer, chat_template=chat_template, bos_token_id=model_obj.token_bos())


class LlamaCppEngine(Engine):
    """The core class that runs inference using llama.cpp."""

    def __init__(
        self,
        model,
        compute_log_probs,
        chat_template=None,
        enable_backtrack=True,
        enable_ff_tokens=True,
        enable_monitoring=True,
        **kwargs,
    ):
        if not is_llama_cpp:
            raise Exception(
                "Please install llama-cpp-python with `pip install llama-cpp-python` in order to use guidance.models.LlamaCpp!"
            )

        if isinstance(model, Path):
            model = str(model)
        if model is None or isinstance(model, str) and len(model.strip()) == 0:
            model = os.environ.get("LLAMA_CPP_MODEL", "")
            if len(model.strip()) == 0:
                try:
                    with open(os.path.expanduser("~/.llama_cpp_model"), "r") as file:
                        model = file.read().replace("\n", "")
                except FileNotFoundError:
                    pass
                except PermissionError:
                    pass
                if len(model.strip()) == 0:
                    raise ValueError(
                        "If model is None then a model file must be specified in either the LLAMA_CPP_MODEL environment variable or in the ~/.llama_cpp_model file."
                    )

        if isinstance(model, str):
            self.model = model
            if "verbose" not in kwargs:
                kwargs["verbose"] = False

            # patch over https://github.com/abetlen/llama-cpp-python/issues/729
            try:
                sys.stdout.fileno()
            except:  # noqa BLE001
                logger.warning(
                    "Cannot use verbose=True in this context (probably CoLab). See https://github.com/abetlen/llama-cpp-python/issues/729"
                )
                kwargs["verbose"] = True  # llama-cpp-python can't hide output in this case

            with normalize_notebook_stdout_stderr():
                self.model_obj = llama_cpp.Llama(model_path=model, logits_all=True, **kwargs)
        elif isinstance(model, llama_cpp.Llama):
            self.model = model.__class__.__name__
            self.model_obj = model
        else:
            raise TypeError("model must be None, a file path string, or a llama_cpp.Llama object.")

        self._context = _LlamaBatchContext(self.model_obj.n_batch, self.model_obj.n_ctx())
        self._n_vocab = self.model_obj.n_vocab()
        self._cached_token_ids = []
        self._cached_logits = None

        super().__init__(
            LlamaCppTokenizer(self.model_obj, chat_template=chat_template),
            compute_log_probs=compute_log_probs,
            enable_backtrack=enable_backtrack,
            enable_ff_tokens=enable_ff_tokens,
            enable_monitoring=enable_monitoring,
            **kwargs,
        )

    def get_logits(self, token_ids: list[int], include_all_uncached_tokens: bool = False) -> LogitsOutput:
        """Computes the logits for the given token state.

        This overrides a method from the LocalEngine class that is used to get
        inference results from the model.
        """

        if len(token_ids) == 0:
            raise ValueError("token_ids must contain some tokens.")

        # make sure we don't run off the end of the model's context
        if self.model_obj.n_ctx() <= len(token_ids):
            raise Exception(
                f"Attempted to use a context length of {len(token_ids)} tokens, but this LlamaCpp model is only configured to support up to {self.model_obj.n_ctx()}!"
            )

        # check what we have already cached
        num_cached = sum(takewhile(operator.truth, map(operator.eq, token_ids, self._cached_token_ids)))
        if num_cached == len(token_ids):
            if num_cached == len(self._cached_token_ids):
                # last token input is the same as the last cached token, so return the last cached logits
                return {
                    "logits": self._cached_logits,
                    "n_tokens": len(token_ids),
                    "n_cached": num_cached,
                }
            # we need to pass at least one new token
            num_cached = num_cached - 1

        # clear obsolete parts of kv cache
        kv = llama_cpp.llama_get_memory(self.model_obj.ctx)
        llama_cpp.llama_memory_seq_rm(kv, -1, num_cached, -1)

        # eval the model
        logits_for_each_batch: list[np.ndarray] = []
        n_batch = self.model_obj.n_batch
        batch = self._context.batch
        for i in range(num_cached, len(token_ids), n_batch):
            n_tokens = min(i + n_batch, len(token_ids)) - i
            batch.n_tokens = n_tokens
            for j in range(n_tokens):
                batch.token[j] = token_ids[i + j]
                batch.pos[j] = i + j
                batch.seq_id[j][0] = 0
                batch.n_seq_id[j] = 1
                batch.logits[j] = True

            ret = llama_cpp.llama_decode(self.model_obj.ctx, batch)
            if ret != 0:
                raise Exception(f"Call to llama_cpp.llama_decode returned {ret}.")

            # get the logits for *this* batch
            llama_logits = llama_cpp.llama_get_logits(self.model_obj.ctx)
            logits_for_this_batch = np.ctypeslib.as_array(
                llama_logits,
                shape=(n_tokens, self._n_vocab),
            ).copy()
            logits_for_each_batch.append(logits_for_this_batch)

        # save the results
        self._cached_token_ids = token_ids.copy()
        self._cached_logits = logits_for_each_batch[-1][[-1], :]

        if include_all_uncached_tokens:
            logits = np.concatenate(logits_for_each_batch, axis=0)
            assert logits.shape[0] == len(token_ids) - num_cached
        else:
            logits = self._cached_logits

        return {
            "logits": logits,
            "n_tokens": len(token_ids),
            "n_cached": num_cached,
        }


class LlamaCpp(Model):
    def __init__(
        self,
        model=None,
        echo=True,
        compute_log_probs=False,
        chat_template=None,
        enable_backtrack=True,
        enable_ff_tokens=True,
        enable_monitoring=True,
        sampling_params: Optional[SamplingParams] = None,
        **llama_cpp_kwargs,
    ):
        """Build a new LlamaCpp model object that represents a model in a given state."""

        engine = LlamaCppEngine(
            model,
            compute_log_probs=compute_log_probs,
            chat_template=chat_template,
            enable_backtrack=enable_backtrack,
            enable_ff_tokens=enable_ff_tokens,
            enable_monitoring=enable_monitoring,
            **llama_cpp_kwargs,
        )
        interpreter = EngineInterpreter(engine)
        super().__init__(
            interpreter=interpreter,
            sampling_params=SamplingParams() if sampling_params is None else sampling_params,
            echo=echo,
        )



from ._tokenizer import Tokenizer  # isort:skip
from ._engine import Engine, LogitsOutput
from ._interpreter import EngineInterpreter, Llama3VisionInterpreter, Phi3VisionInterpreter
from ._state import EngineState

__all__ = [
    "Engine",
    "EngineInterpreter",
    "EngineState",
    "Llama3VisionInterpreter",
    "LogitsOutput",
    "Phi3VisionInterpreter",
    "Tokenizer",
]



from typing import Iterator, Optional

from guidance._schema import SamplingParams

from ..._ast import GrammarNode
from ...trace import OutputAttr, TextOutput
from .._base import Model
from .._openai_base import BaseOpenAIInterpreter, OpenAIClientWrapper


class VLLMInterpreter(BaseOpenAIInterpreter):
    def __init__(
        self,
        model: str,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        **kwargs,
    ):
        try:
            import openai
        except ImportError as ie:
            raise Exception(
                "Please install the openai package version >= 1 using `pip install openai -U` in order to use guidance.models.OpenAI!"
            ) from ie

        client = openai.OpenAI(base_url=base_url, api_key=api_key, **kwargs)
        super().__init__(model=model, client=OpenAIClientWrapper(client), **kwargs)

    def grammar(self, node: GrammarNode, **kwargs) -> Iterator[OutputAttr]:
        buffer: str = ""

        kwargs = self._process_kwargs(**kwargs)
        extra_body = {
            "guided_decoding_backend": "guidance",
            "guided_grammar": node.ll_grammar(),
        }
        kwargs["extra_body"].update(extra_body)

        for attr in self._run(**kwargs):
            if isinstance(attr, TextOutput):
                buffer += attr.value
            yield attr
        matches = node.match(
            buffer,
            raise_exceptions=False,
            # Turn of max_tokens since we don't have access to the tokenizer
            enforce_max_tokens=False,
        )
        if matches is None:
            # TODO: should probably raise...
            # raise ValueError("vLLM failed to constrain the grammar")
            pass
        else:
            for name, value in matches.captures.items():
                log_probs = matches.log_probs[name]
                if isinstance(value, list):
                    assert isinstance(log_probs, list)
                    assert len(value) == len(log_probs)
                    for v, l in zip(value, log_probs):
                        yield self.state.apply_capture(name=name, value=v, log_prob=l, is_append=True)
                else:
                    yield self.state.apply_capture(name=name, value=value, log_prob=log_probs, is_append=False)

    def _process_kwargs(self, **kwargs):
        if "extra_body" not in kwargs:
            kwargs["extra_body"] = {}

        sampling_params = kwargs.pop("sampling_params", None)
        if sampling_params is None:
            return kwargs

        kwargs["top_p"] = sampling_params.pop("top_p", None)

        # top_k must be put in extra_body
        top_k = sampling_params.pop("top_k", None)
        if top_k is not None:
            kwargs["extra_body"]["top_k"] = top_k

        min_p = sampling_params.pop("min_p", None)
        if min_p is not None:
            kwargs["extra_body"]["min_p"] = min_p

        repetition_penalty = sampling_params.pop("repetition_penalty", None)
        if repetition_penalty is not None:
            kwargs["extra_body"]["repetition_penalty"] = repetition_penalty

        return kwargs


class VLLMModel(Model):
    def __init__(self, model: str, sampling_params: Optional[SamplingParams] = None, echo: bool = True, **kwargs):
        super().__init__(
            interpreter=VLLMInterpreter(model=model, **kwargs),
            sampling_params=SamplingParams() if sampling_params is None else sampling_params,
            echo=echo,
        )



import numpy as np

from ._engine import Tokenizer
from ._engine._tokenizer import TokenizerWrappable


class ByteTokenizer(Tokenizer):
    def __init__(self, chat_template=None):
        # directly map integer values to byte strings
        all_bytes = [bytes([i]) for i in range(256)]
        bos = b"<s>"
        tokens = np.array(all_bytes + [bos], dtype="object")
        ll_tokenizer = TokenizerWrappable(
            eos_token_id=256,
            bos_token_id=256,
            tokens=tokens,
            special_token_ids=[],
            # ENCODE MUST BE OVERRIDDEN
            encode_callable=self.encode,
        ).as_ll_tokenizer()

        super().__init__(
            ll_tokenizer=ll_tokenizer,
            chat_template=chat_template,
            bos_token_id=256,
        )

    def encode(self, byte_string: bytes, *, parse_special: bool = True) -> list[int]:
        """Returns a list of tokens that represent the given byte string."""
        if isinstance(byte_string, str):
            byte_string = byte_string.encode("utf8")
        i = 0
        result = []
        while i < len(byte_string):
            if parse_special and byte_string[i : i + 3] == b"<s>":
                result.append(256)
                i += 3  # Skip the next two characters as part of '<s>'
            else:
                result.append(byte_string[i])
                i += 1
        return result



# TODO(nopdive): Consider integrating token operations into trace nodes (handles token healing cleaner).
# TODO(nopdive): Benchmark (expected heap fragmentation issue). Likely need memory pooling (via rust/ctypes/Cython).
import logging
import weakref
from itertools import count
from typing import Annotated, Any, ClassVar, Generator, Optional, Union

from pydantic import Base64Bytes, BaseModel, Discriminator, Field, Tag, computed_field, model_validator

from .._utils import log_cleanup, pydantic_no_default_repr, pydantic_no_default_str

logger = logging.getLogger(__name__)


class NodeAttr(BaseModel):
    """Attributes of a trace node."""

    _subclasses: ClassVar[set[type["NodeAttr"]]] = set()

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._subclasses.add(cls)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def class_name(self) -> str:
        """Class name of the message."""
        return self.__class__.__name__

    @model_validator(mode="before")
    def validate_class_name(cls, data):
        if isinstance(data, dict):
            if "class_name" in data and data["class_name"] != cls.__name__:
                raise ValueError(f"mismatched class name: {data['class_name']}, expected: {cls.__name__}")
        return data

    @classmethod
    def as_discriminated_union(cls) -> type["NodeAttr"]:
        return Annotated[
            Union[tuple(Annotated[tp, Tag(tp.__name__)] for tp in cls._subclasses)],
            Discriminator(
                lambda x: x["class_name"] if isinstance(x, dict) else x.class_name,
            ),
        ]  # type: ignore[return-value]

    def __repr__(self):
        return pydantic_no_default_repr(self)

    def __str__(self):
        return pydantic_no_default_str(self)


class InputAttr(NodeAttr):
    """Input for a guidance program (i.e. literal or guidance grammar)."""

    pass


class OutputAttr(NodeAttr):
    """Output for a guidance program (i.e. text output)."""

    pass


class StatelessGuidanceInput(InputAttr):
    """Stateless guidance input (light wrapper)."""

    # NOTE(nopdive): Open to debate what we should serialize here, excluding for now.
    value: Any = Field(exclude=True)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.value})"


class StatefulGuidanceInput(InputAttr):
    """Stateful guidance input (light wrapper)."""

    # NOTE(nopdive): Open to debate what we should serialize here, excluding for now.
    value: Any = Field(exclude=True)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.value})"


class LiteralInput(InputAttr):
    """Text string as a literal."""

    value: str


class ImageInput(InputAttr):
    """Image input."""

    value: Base64Bytes
    format: str = "png"


class AudioInput(InputAttr):
    """Audio input."""

    value: Base64Bytes
    format: str = "wav"


class VideoInput(InputAttr):
    """Video input."""

    value: Base64Bytes
    format: str = "mp4"


class EmbeddedInput(InputAttr):
    """Text string with embedded guidance input."""

    value: str


class RoleOpenerInput(InputAttr):
    """Appears when a role is opened (i.e. user / system).

    This usually occurs as a role context and __enter__ is called.
    """

    name: Optional[str] = None
    text: Optional[str] = None
    closer_text: Optional[str] = None


class RoleCloserInput(InputAttr):
    """Appears when a role is closed (i.e. user / system).

    This usually occurs as a role context and __exit__ is called.
    """

    name: Optional[str] = None
    text: Optional[str] = None


class AudioOutput(OutputAttr):
    """Audio output."""

    value: Base64Bytes
    format: str = "wav"
    is_input: bool = False


class VideoOutput(OutputAttr):
    """Video output."""

    value: Base64Bytes
    format: str = "mp4"
    is_input: bool = False


class ImageOutput(OutputAttr):
    """Image output."""

    value: Base64Bytes
    format: str = "png"
    is_input: bool = False


class TextOutput(OutputAttr):
    """Text string."""

    value: str
    is_input: bool = False
    is_generated: bool = False
    is_force_forwarded: bool = False
    latency_ms: float = 0.0

    def __str__(self):
        return self.value


class Token(BaseModel):
    token: str
    bytes: Base64Bytes
    prob: float = float("nan")
    masked: bool = False


class TokenOutput(TextOutput):
    token: Token
    top_k: Optional[list[Token]] = None


class Backtrack(OutputAttr):
    n_tokens: int
    bytes: Base64Bytes


class CaptureOutput(OutputAttr):
    """Capture variable output as a string.

    If `value` is set to None, this means it's a reset (needed for append capture group outputs).
    """

    name: str
    value: Optional[str] = None
    is_append: bool = False
    log_probs: float = 0.0

    def __str__(self):
        return f"{self.name}{'+=' if self.is_append else '='}{self.value.__str__()}"


class WeakRefList(list):
    """Weak reference list implementation that uses weakref ref objects.

    This does not override all methods for list.
    """

    def append(self, item):
        super().append(weakref.ref(item))

    def __getitem__(self, index):
        ref = super().__getitem__(index)
        obj = ref()
        if obj is None:
            raise ReferenceError("The referenced object has been garbage collected")
        return obj

    def __iter__(self):
        return (ref() for ref in super().__iter__() if ref() is not None)

    def remove(self, item):
        ref_to_remove = None
        for ref in super().__iter__():
            obj = ref()
            if obj is item:
                ref_to_remove = ref
                break
        if ref_to_remove:
            super().remove(ref_to_remove)


def _cleanup(log_msg: str):
    log_cleanup(log_msg)


class TraceNode(BaseModel):
    """Trace node which associates inputs and outputs of a guidance program."""

    identifier: int = Field(default_factory=count().__next__)
    parent: Optional["TraceNode"] = None
    children: list["TraceNode"] = Field(default_factory=WeakRefList)
    input: list[InputAttr] = Field(default_factory=list)
    output: list[OutputAttr] = Field(default_factory=list)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        weakref.finalize(self, _cleanup, f"trace({id(self)}):{self.identifier}")

    def add_child(self, child: "TraceNode") -> None:
        """Add a child node to the trace node.

        Args:
            child: The child node to add.
        """
        child.parent = self
        self.children.append(child)

    def remove_child(self, child: "TraceNode") -> None:
        """Remove a child node from the trace node.

        Args:
            child: The child node to remove.
        """
        child.parent = None
        self.children.remove(child)

    def ancestors(self) -> Generator["TraceNode", None, None]:
        """Yields ancestor trace nodes ending with root.

        Yields:
            Ancestor trace nodes.
        """
        node = self
        while node.parent is not None:
            yield node.parent
            node = node.parent

    def path(self) -> Generator["TraceNode", None, None]:
        """Yields path of from root to self inclusively.

        Yields:
            Trace nodes from root to self.
        """
        yield from reversed(list(self.ancestors()))
        yield self

    def root(self) -> "TraceNode":
        """Returns root by traversing parents of self.

        Returns:
            Root of tree self is part of.
        """
        if self.parent is None:
            root = self
        else:
            *_, root = self.ancestors()
        return root

    def traverse(self, bfs: bool = True):
        """Traverse the trace nodes starting with self.

        Args:
            bfs: Use breadth-first-search, otherwise depth-first-search.

        Yields:
            Trace nodes in traversal order.
        """
        queue = [self]
        while queue:
            node = queue.pop(0)
            yield node

            if bfs:
                queue.extend(node.children)
            else:
                # NOTE(nopdive): Analogous to extend but at front of list.
                queue[0:0] = node.children

    def __repr__(self):
        return f"{self.identifier}:{self.input!r}:{self.output!r}"

    def __hash__(self):
        return hash(self.identifier)


class TraceHandler(BaseModel):
    """Trace handler that will own a tree of trace nodes.

    This will primarily be owned by a model's engine.
    Each guidance model corresponds to one trace node.
    All guidance models emitted from an engine is to be included as their own paths within the tree.

    The requirement for holding all live traces ensures downstream consumers such as UI providers
    can do near-real-time partial updates.
    """

    # NOTE(nopdive): Type trickery for pydantic.
    id_node_map: dict[int, TraceNode] = weakref.WeakValueDictionary()  # type: ignore
    node_id_map: dict[TraceNode, int] = weakref.WeakKeyDictionary()  # type: ignore

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        weakref.finalize(self, _cleanup, f"tracehandler({id(self)})")

    def __getitem__(self, item):
        return self.id_node_map[item]

    def __hash__(self):
        return hash(id(self))

    def update_node(self, identifier: int, parent_id: Optional[int], node_attr: Optional[NodeAttr] = None) -> TraceNode:
        """Update the trace node with the given identifier.

        If the trace node does not exist, it will be created.
        Both parent id and node attributes can be updated only once until further notice.

        Args:
            identifier: User-defined identifier for the trace node.
            parent_id: User-defined parent identifier for the trace node.
            node_attr: Input or output node attribute to be updated on the trace node.
        """
        node = self.id_node_map.get(identifier, None)
        if node is None:
            node = TraceNode()

            self.id_node_map[identifier] = node
            self.node_id_map[node] = identifier

            if parent_id is not None:
                parent_node = self.id_node_map.get(parent_id, None)
                if parent_node is not None:
                    parent_node.add_child(node)

        if node_attr is not None:
            if isinstance(node_attr, InputAttr):
                if node.input:
                    logger.debug(
                        f"Adding additional input to trace node {node.identifier}, now has {len(node.input) + 1} inputs"
                    )
                node.input.append(node_attr)
            elif isinstance(node_attr, OutputAttr):
                if node.output:
                    logger.debug(
                        f"Adding additional output to trace node {node.identifier}, now has {len(node.output) + 1} outputs"
                    )
                node.output.append(node_attr)
            else:
                raise ValueError(f"Unexpected node attr: {node_attr}")
        return node

    def root(self) -> TraceNode:
        """Returns root node of trace handler.

        Raises:
            Exception: If root cannot be found.

        Returns: Root trace node.
        """
        root = None
        for _, node in self.id_node_map.items():
            if node.parent is None:
                root = node
                break
        if root is None:
            raise Exception("No root in trace handler.")
        return root



from dataclasses import dataclass
from functools import cached_property
from typing import Any, Callable, Optional, Sequence, Union

import llguidance

from ...chat import ChatTemplate, load_template_class


@dataclass
class TokenizerWrappable:
    eos_token_id: int
    bos_token_id: Optional[int]
    tokens: list[bytes]
    special_token_ids: list[int]
    encode_callable: Callable[[bytes], list[int]]

    def __call__(self, byte_string: bytes) -> list[int]:
        return self.encode_callable(byte_string)

    def as_ll_tokenizer(self) -> "llguidance.LLTokenizer":
        """Returns an LLTokenizer that can be used by llguidance."""
        return llguidance.LLTokenizer(llguidance.TokenizerWrapper(self))


class Tokenizer:
    """This is the standardized tokenizer interface used by guidance models.

    This class should be subclassed by specific implementations and then used as the
    tokenizer in the corresponding Engine subclass.
    """

    def __init__(
        self,
        ll_tokenizer: llguidance.LLTokenizer,
        chat_template: Union[str, ChatTemplate, None],
        bos_token_id: Optional[int] = None,
    ):
        self._ll_tokenizer = ll_tokenizer
        # This method supports None, a huggingface style jinja2_template_str, or a ChatTemplate subclass
        # Defaults to ChatML if nothing is found
        self._chat_template = load_template_class(chat_template)
        self._bos_token_id = bos_token_id

    def is_special_token(self, token_id: int) -> bool:
        """Returns True if the given token ID is a special token."""
        return self._ll_tokenizer.is_special_token(token_id)

    @property
    def bos_token_id(self) -> Union[int, None]:
        # Currently, lltokenizer does not have a bos_token attribute,
        # so we have to store our own if we want to use it
        return self._bos_token_id

    @property
    def eos_token_id(self) -> int:
        return self._ll_tokenizer.eos_token

    @cached_property
    def bos_token(self) -> Union[bytes, None]:
        if self.bos_token_id is None:
            return None
        return self.decode([self.bos_token_id])

    @cached_property
    def eos_token(self) -> bytes:
        return self.decode([self.eos_token_id])

    @property
    def chat_template(self) -> Union[Any, None]:
        return self._chat_template

    def __call__(self, byte_string: bytes):
        return self.encode(byte_string)

    def encode(self, byte_string: bytes, *, parse_special: bool = True) -> list[int]:
        """Returns a list of tokens that represent the given byte string."""
        return self._ll_tokenizer.tokenize_bytes(byte_string, parse_special=parse_special)

    def decode(self, tokens: Sequence[int]) -> bytes:
        """Returns the bytes represented by the given list of tokens."""
        return self._ll_tokenizer.decode_bytes(list(tokens))

    def recode(self, tokens: Sequence[int]) -> list[int]:
        """Redoes a tokenization.

        Encoding a string into tokens does not distribute over concatenation.
        That is, in general, `encode(A)+encode(B) != encode(A+B)` (although it
        it may in some cases).
        An LLM will generate token-by-token, but it is possible (even likely) that
        when the generation is considered as a whole, a better tokenization may
        be possible.
        This method takes in a sequence of tokens, and returns an 'improved' sequence.
        """

        # This is the notional behavior
        # It may need to be overridden in particular cases because
        # we are dealing with LLM ecosystems in the real world
        return self.encode(self.decode(tokens))



import logging
from typing import TYPE_CHECKING, Any, Callable, ContextManager, Iterator, Optional, Union, cast

from guidance._schema import SamplingParams

from .._ast import (
    JsonNode,
)
from ..trace import OutputAttr
from ._base import Model
from ._openai_base import (
    BaseOpenAIClientWrapper,
    BaseOpenAIInterpreter,
    OpenAIAudioMixin,
    OpenAIClientWrapper,
    OpenAIImageMixin,
    OpenAIJSONMixin,
    OpenAIRegexMixin,
    OpenAIRuleMixin,
)

if TYPE_CHECKING:
    import azure.ai.inference
    from azure.core.credentials import AzureKeyCredential, TokenCredential
    from openai.types.chat import ChatCompletionChunk

logger = logging.getLogger(__name__)


class AzureOpenAIInterpreter(OpenAIRuleMixin, OpenAIJSONMixin, OpenAIRegexMixin, BaseOpenAIInterpreter):
    """A basic class for interacting with Azure OpenAI."""

    def __init__(
        self,
        *,
        azure_endpoint: str,
        azure_deployment: str,
        model_name: str,
        api_version: Optional[str] = None,
        api_key: Optional[str] = None,
        azure_ad_token: Optional[str] = None,
        azure_ad_token_provider: Optional[Callable[[], str]] = None,
        **kwargs,
    ):
        try:
            import openai
        except ImportError as ie:
            raise Exception(
                "Please install the openai package version >= 1 using `pip install openai -U` in order to use guidance.models.OpenAI!"
            ) from ie
        client = openai.AzureOpenAI(
            azure_endpoint=azure_endpoint,
            azure_deployment=azure_deployment,
            api_version=api_version,
            api_key=api_key,
            azure_ad_token=azure_ad_token,
            azure_ad_token_provider=azure_ad_token_provider,
            **kwargs,
        )
        super().__init__(model=model_name, client=OpenAIClientWrapper(client))


class AzureOpenAIAudioInterpreter(OpenAIAudioMixin, AzureOpenAIInterpreter):
    """Class to add audio capabilities to an Azure OpenAI model"""

    pass


class AzureOpenAIImageInterpreter(OpenAIImageMixin, AzureOpenAIInterpreter):
    """Class to add image capabilities to an Azure OpenAI model"""

    pass


def create_azure_openai_model(
    azure_endpoint: str,
    azure_deployment: str,
    echo: bool = True,
    *,
    model_name: str,
    api_version: Optional[str] = None,
    api_key: Optional[str] = None,
    azure_ad_token: Optional[str] = None,
    azure_ad_token_provider: Optional[Callable[[], str]] = None,
    has_audio_support: bool = False,
    has_image_support: bool = False,
    sampling_params: Optional[SamplingParams] = None,
    **kwargs,
) -> Model:
    """Create a Model capable of interacting with an Azure AI OpenAI deployment

    Parameters
    ----------
    azure_endpoint : str
        The endpoint which holds the OpenAI model deployment. It will probably be
        https://<AZURE OPENAI RESOURCE NAME>.openai.azure.com/"
    azure_deployment : str
        The Azure deployment name to use for the model. The Azure AI portal will
        default this to being the model_name, but it can be different
    model_name : str
        The name of the Azure OpenAI model to use (e.g. gpt-4o-mini).
    api_version : str | None
        The API version to use for the Azure OpenAI service.
    api_key : str | None
        The API key to use for the Azure OpenAI service.
    azure_ad_token : str | None
        The Azure AD token to use for authentication.
    azure_ad_token_provider : Callable[[], str] | None
        A callable that returns an Azure AD token for authentication.
    echo : bool
        If true the final result of creating this model state will be displayed (as HTML in a notebook).
    has_audio_support : bool
        Indicates if the deployed model has support for audio. This factory attempts
        to work this out from the model_name, but this can be used to force the addition
        of audio support to the returned Model.
    has_image_support : bool
        Indicates if the deployed model has support for images. This factory attempts
        to work this out from the model_name, but this can be used to force the addition
        of image support to the returned Model.
    **kwargs :
        All extra keyword arguments are passed directly to the `openai.OpenAI` constructor. Commonly used argument
        names include `base_url` and `organization`
    """
    if has_audio_support and has_image_support:
        raise ValueError("No known models have both audio and image support")

    interpreter_cls: type[AzureOpenAIInterpreter]
    if (model_name and "audio-preview" in model_name) or has_audio_support:
        interpreter_cls = AzureOpenAIAudioInterpreter
    elif (model_name and (model_name.startswith("gpt-4o") or model_name.startswith("o1"))) or has_image_support:
        interpreter_cls = AzureOpenAIImageInterpreter
    else:
        interpreter_cls = AzureOpenAIInterpreter

    interpreter = interpreter_cls(
        azure_endpoint=azure_endpoint,
        model_name=model_name,
        azure_deployment=azure_deployment,
        api_version=api_version,
        api_key=api_key,
        azure_ad_token=azure_ad_token,
        azure_ad_token_provider=azure_ad_token_provider,
        **kwargs,
    )

    model = Model(
        interpreter=interpreter,
        echo=echo,
        sampling_params=SamplingParams() if sampling_params is None else sampling_params,
    )

    return model


class AzureAIClientWrapper(BaseOpenAIClientWrapper):
    def __init__(self, client: "azure.ai.inference.ChatCompletionsClient"):
        self.client = client

    def streaming_chat_completions(
        self,
        model: str,
        messages: list[dict[str, Any]],
        logprobs: bool,
        **kwargs,
    ) -> ContextManager[Iterator["ChatCompletionChunk"]]:
        request = self.client.complete(
            body={
                "model": model,
                "messages": messages,
                "logprobs": logprobs,
                "stream": True,
                **kwargs,
            },
            headers={
                "extra-parameters": "pass-through",
            },
        )
        # It's at least... "mostly" compliant with the OpenAI API?
        return cast(
            ContextManager[Iterator["ChatCompletionChunk"]],
            request,
        )


class AzureInferenceInterpreter(OpenAIRuleMixin, OpenAIJSONMixin, OpenAIRegexMixin, BaseOpenAIInterpreter):
    def __init__(
        self,
        *,
        endpoint: str,
        credential: Union["AzureKeyCredential", "TokenCredential"],
        model_name: str,
    ):
        try:
            import azure.ai.inference
        except ImportError as ie:
            raise Exception(
                "Please install the azure-ai-inference package  using `pip install azure-ai-inference` in order to use guidance.models.AzureInference!"
            ) from ie
        client = azure.ai.inference.ChatCompletionsClient(
            endpoint=endpoint,
            credential=credential,
        )
        super().__init__(model=model_name, client=AzureAIClientWrapper(client))

    def json(self, node: JsonNode, **kwargs) -> Iterator[OutputAttr]:
        return self._run(
            json_schema={
                "name": "json_schema",  # TODO?
                "schema": node.schema,
                "strict": True,
            },
            **kwargs,
        )


def create_azure_aifoundry_model(
    azure_endpoint: str,
    echo: bool = True,
    *,
    model_name: str,
    api_key: Optional[str] = None,
    token_credential: Optional["TokenCredential"] = None,
    sampling_params: Optional[SamplingParams] = None,
) -> Model:
    """Create a Model capable of interacting with an Azure AI OpenAI deployment

    Parameters
    ----------
    azure_endpoint : str
        The endpoint which holds the OpenAI model deployment. It can be obtained
        from the AI Foundry portal, and will look something like
        "https://<DEPLOYMENT_NAME>.<AZURE REGION>.models.ai.azure.com"
    echo : bool
        If true the final result of creating this model state will be displayed (as HTML in a notebook).
    model_name : str
        The actual name of the deployed model
    api_key : str
        One of the authentication options. This can be obtained from the AI Foundry portal
    token_credential :
        The other authentication option. An Azure Token Credential
    """
    try:
        from azure.core.credentials import AzureKeyCredential, TokenCredential
    except ImportError as ie:
        raise Exception(
            "Please install the azure-core package using `pip install -U azure-core` in order to use guidance.models.AzureAI!"
        ) from ie

    credential: Union[AzureKeyCredential, TokenCredential, None] = None
    if api_key and token_credential:
        raise ValueError("Specify either api_key or token_credential")
    elif api_key:
        credential = AzureKeyCredential(api_key)
    elif token_credential:
        credential = token_credential
    else:
        raise ValueError("Must specify either api_key or token_credential")

    interpreter = AzureInferenceInterpreter(
        endpoint=azure_endpoint,
        credential=credential,
        model_name=model_name,
    )

    result = Model(
        interpreter=interpreter,
        echo=echo,
        sampling_params=SamplingParams() if sampling_params is None else sampling_params,
    )
    return result



"""Background thread for asyncio handling.

This is currently being used for messaging, visualization and metrics.
"""

import asyncio
import threading
from asyncio import AbstractEventLoop, Task
from concurrent.futures import Future
from typing import Any, Coroutine, TypeVar

T = TypeVar("T")


def _start_asyncio_loop(loop: AbstractEventLoop):
    asyncio.set_event_loop(loop)
    loop.run_forever()


def _asyncio_background_thread() -> tuple[threading.Thread, AbstractEventLoop]:
    loop = asyncio.new_event_loop()
    thread = threading.Thread(target=_start_asyncio_loop, args=(loop,))
    thread.daemon = True
    return thread, loop


class BackgroundAsync:
    """Runs background thread that has an asyncio event loop."""

    def __init__(self):
        """Initializes."""
        self._loop = None
        self._thread = None

    def _thread_and_loop(self) -> tuple[threading.Thread, AbstractEventLoop]:
        if self._loop is None:
            self._thread, self._loop = _asyncio_background_thread()
            self._thread.start()
        return self._thread, self._loop

    def call_soon_threadsafe(self, cb, *args, context=None):
        """Fires callback in background thread."""

        _, loop = self._thread_and_loop()
        return loop.call_soon_threadsafe(cb, *args, context=context)

    def run_async_coroutine(self, coroutine: Coroutine[Any, Any, T]) -> Future[T]:
        """Runs an asynchronous coroutine in the visual thread.

        Args:
            coroutine: Coroutine to be run on visual thread.

        Returns:
            Future of coroutine.
        """
        _, loop = self._thread_and_loop()
        future = asyncio.run_coroutine_threadsafe(coroutine, loop)
        return future

    @staticmethod
    async def async_task(coroutine: Coroutine[Any, Any, T]) -> Task[T]:
        """Creates an asyncio task from coroutine.

        Args:
            coroutine: Coroutine within task.

        Returns:
            Asyncio task.
        """
        task = asyncio.create_task(coroutine)
        return task

    @staticmethod
    async def print_all_tasks():  # pragma: no cover
        """Prints all tasks running in background thread loop (for debugging purposes)."""
        for task in asyncio.all_tasks():
            print(task)



"""Messages that used between server (usually Jupyter Python kernel) and client.

Messages are required to be added to the model registry for serialization.
"""

from itertools import count
from typing import Annotated, ClassVar, Optional, Union

from pydantic import BaseModel, Discriminator, Field, Tag, TypeAdapter, computed_field, model_validator

from ..trace import NodeAttr


class GuidanceMessage(BaseModel):
    """Message sent within Guidance layer."""

    message_id: int = Field(default_factory=count().__next__)

    _subclasses: ClassVar[set[type["GuidanceMessage"]]] = set()

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._subclasses.add(cls)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def class_name(self) -> str:
        """Class name of the message."""
        return self.__class__.__name__

    @model_validator(mode="before")
    def validate_class_name(cls, data):
        if isinstance(data, dict):
            if "class_name" in data and data["class_name"] != cls.__name__:
                raise ValueError(f"mismatched class name: {data['class_name']}, expected: {cls.__name__}")
        return data

    @classmethod
    def as_discriminated_union(cls) -> type["GuidanceMessage"]:
        return Annotated[
            Union[tuple(Annotated[tp, Tag(tp.__name__)] for tp in cls._subclasses)],
            Discriminator(
                lambda x: x["class_name"] if isinstance(x, dict) else x.class_name,
            ),
        ]  # type: ignore[return-value]


class TraceMessage(GuidanceMessage):
    """Update on a trace node."""

    trace_id: int
    parent_trace_id: Optional[int] = None
    node_attr: Optional[NodeAttr.as_discriminated_union()] = None  # type: ignore


class MetricMessage(GuidanceMessage):
    """Metric that has been emitted."""

    name: str
    value: Union[float, str, list[float], list[str]] = Field(union_mode="left_to_right")
    scalar: bool = True


class ExecutionStartedMessage(GuidanceMessage):
    """Fired when renderer has started trace messages."""


class ExecutionCompletedMessage(GuidanceMessage):
    """Fired when renderer has completed trace messages.

    This functions as the last message sent to client.
    """

    last_trace_id: Optional[int] = None
    is_err: bool = False


class ResetDisplayMessage(GuidanceMessage):
    """Instructs client to reset the display, removing all output."""

    pass


class ClientReadyMessage(GuidanceMessage):
    """Fired when client is ready to receive messages."""

    pass


class ClientReadyAckMessage(GuidanceMessage):
    """Fired when server acknowledges client readiness."""

    pass


class OutputRequestMessage(GuidanceMessage):
    """Fired when client requests tokens from server."""

    identifier: str = ""


class OutputRequestAckMessage(GuidanceMessage):
    """Fired when server acknowledges request."""

    pass


def serialize_message(message: GuidanceMessage) -> str:
    """Serializes guidance message.

    Args:
        message: Message to be serialized.

    Returns:
        Serialized message in JSON format.
    """
    return message.model_dump_json()


def deserialize_message(data: str) -> GuidanceMessage:
    """Deserializes string into a guidance message.

    Args:
        data: JSON string to be deserialized.

    Returns:
        Guidance message.
    """
    return TypeAdapter(GuidanceMessage.as_discriminated_union()).validate_json(data)
